Package: GenomicDataCommons
Type: Package
Title: NIH / NCI Genomic Data Commons Access
Description: Programmatically access the NIH / NCI Genomic Data Commons
        RESTful service.
Version: 1.35.1
Date: 2025-05-12
Authors@R: c( person("Martin", "Morgan",
        email="martin.morgan@roswellpark.org", role=c("aut")),
        person("Sean", "Davis", email="seandavi@gmail.com",
        role=c("aut", "cre")),
        person("Marcel", "Ramos",
        email = "marcel.ramos@sph.cuny.edu", role = "ctb"))
License: Artistic-2.0
Depends: R (>= 4.1.0)
Imports: stats, httr, xml2, jsonlite, utils, rlang, readr,
        GenomicRanges, IRanges, dplyr, rappdirs, tibble, tidyr
Suggests: BiocStyle, knitr, rmarkdown, DT, testthat, listviewer,
        ggplot2, GenomicAlignments, Rsamtools, BiocParallel,
        TxDb.Hsapiens.UCSC.hg38.knownGene, VariantAnnotation, maftools,
        R.utils, data.table
biocViews: DataImport, Sequencing
URL: https://bioconductor.org/packages/GenomicDataCommons,
        http://github.com/Bioconductor/GenomicDataCommons,
        http://bioconductor.github.io/GenomicDataCommons/
BugReports: https://github.com/Bioconductor/GenomicDataCommons/issues/new
Encoding: UTF-8
VignetteBuilder: knitr
RoxygenNote: 7.3.2
git_url: https://git.bioconductor.org/packages/GenomicDataCommons
git_branch: devel
git_last_commit: 06edbb7
git_last_commit_date: 2025-11-07
Repository: Bioconductor 3.23
Date/Publication: 2025-11-07
NeedsCompilation: no
Packaged: 2025-11-07 22:44:25 UTC; biocbuild
Author: Martin Morgan [aut],
  Sean Davis [aut, cre],
  Marcel Ramos [ctb]
Maintainer: Sean Davis <seandavi@gmail.com>

# Generated by roxygen2: do not edit by hand

S3method(aggregations,GDCQuery)
S3method(aggregations,GDCResponse)
S3method(available_expand,GDCQuery)
S3method(available_expand,character)
S3method(available_fields,GDCQuery)
S3method(available_fields,character)
S3method(count,GDCQuery)
S3method(count,GDCResponse)
S3method(default_fields,GDCQuery)
S3method(default_fields,character)
S3method(entity_name,GDCQuery)
S3method(entity_name,GDCResults)
S3method(expand,GDCQuery)
S3method(facet,GDCQuery)
S3method(field_description,GDCQuery)
S3method(field_description,character)
S3method(filter,GDCQuery)
S3method(get_facets,GDCQuery)
S3method(get_filter,GDCQuery)
S3method(id_field,GDCQuery)
S3method(id_field,GDCResults)
S3method(ids,GDCManifest)
S3method(ids,GDCQuery)
S3method(ids,GDCResponse)
S3method(ids,GDCResults)
S3method(manifest,GDCcasesResponse)
S3method(manifest,GDCfilesResponse)
S3method(manifest,gdc_files)
S3method(print,gdc_endpoints)
S3method(print,gdc_parameters)
S3method(response,GDCQuery)
S3method(results,GDCQuery)
S3method(results,GDCResponse)
S3method(results_all,GDCQuery)
S3method(results_all,GDCResponse)
S3method(select,GDCQuery)
export(aggregations)
export(annotations)
export(available_expand)
export(available_fields)
export(available_values)
export(cases)
export(cnv_occurrences)
export(cnvs)
export(count)
export(default_fields)
export(endpoints)
export(entity_name)
export(expand)
export(facet)
export(field_description)
export(files)
export(filter)
export(gdc_cache)
export(gdc_client)
export(gdc_client_version_validate)
export(gdc_clinical)
export(gdc_set_cache)
export(gdc_token)
export(gdcdata)
export(genes)
export(get_facets)
export(get_filter)
export(grep_fields)
export(id_field)
export(ids)
export(make_filter)
export(manifest)
export(mapping)
export(parameters)
export(projects)
export(query)
export(readDNAcopy)
export(readHTSeqFile)
export(response)
export(response_all)
export(results)
export(results_all)
export(select)
export(slicing)
export(ssm_occurrences)
export(ssms)
export(status)
export(transfer)
export(transfer_help)
export(write_manifest)
import(GenomicRanges)
importFrom(IRanges,IRanges)
importFrom(dplyr,bind_rows)
importFrom(httr,GET)
importFrom(httr,POST)
importFrom(httr,add_headers)
importFrom(httr,content)
importFrom(httr,headers)
importFrom(httr,progress)
importFrom(httr,stop_for_status)
importFrom(httr,write_disk)
importFrom(jsonlite,fromJSON)
importFrom(jsonlite,toJSON)
importFrom(jsonlite,unbox)
importFrom(rappdirs,app_dir)
importFrom(readr,read_tsv)
importFrom(rlang,enquo)
importFrom(rlang,eval_tidy)
importFrom(rlang,f_env)
importFrom(rlang,f_rhs)
importFrom(rlang,is_formula)
importFrom(stats,setNames)
importFrom(tibble,as_tibble)
importFrom(utils,menu)
importFrom(utils,read.table)
importFrom(utils,write.table)
importFrom(xml2,xml_find_all)
importFrom(xml2,xml_text)

# GenomicDataCommons

<!-- badges: start -->

[![R-CMD-check](https://github.com/Bioconductor/GenomicDataCommons/workflows/R-CMD-check/badge.svg)](https://github.com/Bioconductor/GenomicDataCommons/actions)
<!-- badges: end -->

# What is the GDC?

From the [Genomic Data Commons (GDC)
website](https://gdc.nci.nih.gov/about-gdc):

The National Cancer Institute’s (NCI’s) Genomic Data Commons (GDC) is a
data sharing platform that promotes precision medicine in oncology. It
is not just a database or a tool; it is an expandable knowledge network
supporting the import and standardization of genomic and clinical data
from cancer research programs.

The GDC contains NCI-generated data from some of the largest and most
comprehensive cancer genomic datasets, including The Cancer Genome Atlas
(TCGA) and Therapeutically Applicable Research to Generate Effective
Therapies (TARGET). For the first time, these datasets have been
harmonized using a common set of bioinformatics pipelines, so that the
data can be directly compared.

As a growing knowledge system for cancer, the GDC also enables
researchers to submit data, and harmonizes these data for import into
the GDC. As more researchers add clinical and genomic data to the GDC,
it will become an even more powerful tool for making discoveries about
the molecular basis of cancer that may lead to better care for patients.

The [data model for the GDC is
complex](https://gdc.cancer.gov/developers/gdc-data-model/gdc-data-model-components),
but it worth a quick overview. The data model is encoded as a so-called
property graph. Nodes represent entities such as Projects, Cases,
Diagnoses, Files (various kinds), and Annotations. The relationships
between these entities are maintained as edges. Both nodes and edges may
have Properties that supply instance details. The GDC API exposes these
nodes and edges in a somewhat simplified set of
[RESTful](https://en.wikipedia.org/wiki/Representational_state_transfer)
endpoints.

# Quickstart

This software is available at Bioconductor.org and can be downloaded via
`BiocManager::install`.

To report bugs or problems, either [submit a new
issue](https://github.com/Bioconductor/GenomicDataCommons/issues) or
submit a `bug.report(package='GenomicDataCommons')` from within R (which
will redirect you to the new issue on GitHub).

## Installation

Installation can be achieved via Bioconductor’s `BiocManager` package.

``` r
if (!require("BiocManager"))
    install.packages("BiocManager")

BiocManager::install('GenomicDataCommons')
```

``` r
library(GenomicDataCommons)
```

## Check basic functionality

``` r
status()
#> $commit
#> [1] "4dd3680528a19ed33cfc83c7d049426c97bb903b"
#> 
#> $data_release
#> [1] "Data Release 34.0 - July 27, 2022"
#> 
#> $status
#> [1] "OK"
#> 
#> $tag
#> [1] "3.0.0"
#> 
#> $version
#> [1] 1
```

## Find data

The following code builds a `manifest` that can be used to guide the
download of raw data. Here, filtering finds gene expression files
quantified as raw counts using `STAR` from ovarian cancer patients.

``` r
ge_manifest <- files() |>
    filter( cases.project.project_id == 'TCGA-OV') |>
    filter( type == 'gene_expression' ) |>
    filter( analysis.workflow_type == 'STAR - Counts') |>
    manifest(size = 5)
ge_manifest
#>                                     id data_format     access                                                                   file_name
#> 1 7c69529f-2273-4dc4-b213-e84924d78bea         TSV       open d6472bd0-b4e2-4ed1-a892-e1702c195dc7.rna_seq.augmented_star_gene_counts.tsv
#> 2 0eff4634-f8c4-4db9-8a7c-331b21689bae         TSV       open 42165baf-b32c-4fc4-8b04-29c5b4e76de0.rna_seq.augmented_star_gene_counts.tsv
#> 3 7d74b4c5-6391-4b3e-95a3-020ea0869e86         TSV controlled   accf08d4-a784-4908-831a-7a08d4c5f0f5.rna_seq.star_splice_junctions.tsv.gz
#> 4 dc2aeea4-3cd0-4623-92f4-bbbc962851cc         TSV controlled   8ab508b9-2993-4e66-b8f9-81e32e936d4a.rna_seq.star_splice_junctions.tsv.gz
#> 5 0cf852be-d2e3-4fde-bba8-c93efae2961a         TSV       open 93831282-1dd1-49a3-acd7-dae2a49ca62e.rna_seq.augmented_star_gene_counts.tsv
#>                           submitter_id           data_category       acl            type file_size                 created_datetime                           md5sum
#> 1 7085a70b-2f63-4402-9e53-70f091f26fcb Transcriptome Profiling      open gene_expression   4254435 2021-12-13T20:53:42.329364-06:00 19d5596bba8949f4c138793608497d56
#> 2 f0d44930-b1ad-447a-86b9-27d0285954b9 Transcriptome Profiling      open gene_expression   4257461 2021-12-13T20:47:24.326497-06:00 d89d71b7c028c1643d7a3ee7857d8e01
#> 3 e6473134-6d65-414c-9f52-2c25057fac7d Transcriptome Profiling phs000178 gene_expression   3109435 2021-12-13T21:03:56.008440-06:00 fb8332d6413c44a9de02a1cbe6b018aa
#> 4 f99b93a9-70cb-44f8-bd1f-4edeee4425a4 Transcriptome Profiling phs000178 gene_expression   4607701 2021-12-13T21:02:23.944851-06:00 26231bed1ef67c093d3ce2b39def81cd
#> 5 fb4d7abe-b61a-4f35-9700-605f1bc1512f Transcriptome Profiling      open gene_expression   4265694 2021-12-13T20:50:55.234254-06:00 050763aabd36509f954137fbdc4eeb00
#>                   updated_datetime                              file_id                      data_type    state experimental_strategy
#> 1 2022-01-19T14:47:28.965154-06:00 7c69529f-2273-4dc4-b213-e84924d78bea Gene Expression Quantification released               RNA-Seq
#> 2 2022-01-19T14:47:07.478144-06:00 0eff4634-f8c4-4db9-8a7c-331b21689bae Gene Expression Quantification released               RNA-Seq
#> 3 2022-01-19T14:01:15.621847-06:00 7d74b4c5-6391-4b3e-95a3-020ea0869e86 Splice Junction Quantification released               RNA-Seq
#> 4 2022-01-19T14:01:15.621847-06:00 dc2aeea4-3cd0-4623-92f4-bbbc962851cc Splice Junction Quantification released               RNA-Seq
#> 5 2022-01-19T14:47:07.036781-06:00 0cf852be-d2e3-4fde-bba8-c93efae2961a Gene Expression Quantification released               RNA-Seq
```

## Download data

This code block downloads the 5 gene expression files specified in the
query above. Using multiple processes to do the download very
significantly speeds up the transfer in many cases. The following
completes in about 15 seconds.

``` r
library(BiocParallel)
register(MulticoreParam())
destdir <- tempdir()
fnames <- lapply(ge_manifest$id,gdcdata)
```

If the download had included controlled-access data, the download above
would have needed to include a `token`. Details are available in [the
authentication section below](#authentication).

## Metadata queries

Here we use a couple of ad-hoc helper functions to handle the output of
the query. See the `inst/script/README.Rmd` folder for the source.

First, create a `data.frame` from the clinical data:

``` r
expands <- c("diagnoses","annotations",
             "demographic","exposures")
clinResults <- cases() |>
    GenomicDataCommons::select(NULL) |>
    GenomicDataCommons::expand(expands) |>
    results(size=6)
demoDF <- filterAllNA(clinResults$demographic)
exposuresDF <- bindrowname(clinResults$exposures)
```

``` r
demoDF[, 1:4]
#>                                      cause_of_death         race gender              ethnicity
#> 2525bfef-6962-4b7f-8e80-6186400ce624           <NA> not reported female           not reported
#> 126507c3-c0d7-41fb-9093-7deed5baf431 Cancer Related not reported female           not reported
#> c43ac461-9f03-44bc-be7d-3d867eb708a0           <NA> not reported female           not reported
#> a59a90d9-f1b0-49dd-9c97-bcaa6ba55d44 Cancer Related not reported   male           not reported
#> 59122a43-606a-4669-806b-6747e0ac9985           <NA>        white   male not hispanic or latino
#> 4447a969-e5c8-4291-b83c-53a0f7e77cbc Cancer Related        white female not hispanic or latino
```

``` r
exposuresDF[, 1:4]
#>                                       submitter_id                 created_datetime    alcohol_intensity pack_years_smoked
#> 2525bfef-6962-4b7f-8e80-6186400ce624 C3N-03839-EXP 2019-12-30T10:23:07.190853-06:00 Lifelong Non-Drinker                NA
#> 126507c3-c0d7-41fb-9093-7deed5baf431 C3N-01518-EXP 2018-06-21T14:27:48.817254-05:00 Lifelong Non-Drinker                NA
#> c43ac461-9f03-44bc-be7d-3d867eb708a0 C3N-03933-EXP 2019-03-14T08:23:14.054975-05:00 Lifelong Non-Drinker                NA
#> a59a90d9-f1b0-49dd-9c97-bcaa6ba55d44 C3N-02695-EXP 2019-03-14T08:23:14.054975-05:00   Occasional Drinker              16.8
#> 59122a43-606a-4669-806b-6747e0ac9985 C3L-03642-EXP 2019-06-24T07:53:15.534197-05:00 Lifelong Non-Drinker              39.0
#> 4447a969-e5c8-4291-b83c-53a0f7e77cbc C3L-03728-EXP 2019-06-24T07:53:15.534197-05:00 Lifelong Non-Drinker                NA
```

Note that the diagnoses data has multiple lines per patient:

``` r
diagDF <- bindrowname(clinResults$diagnoses)
diagDF[, 1:4]
#>                                      ajcc_pathologic_stage                 created_datetime tissue_or_organ_of_origin age_at_diagnosis
#> 2525bfef-6962-4b7f-8e80-6186400ce624             Stage IIB 2019-07-22T06:40:02.183501-05:00          Head of pancreas            19956
#> 126507c3-c0d7-41fb-9093-7deed5baf431          Not Reported 2018-12-03T12:05:16.846188-06:00             Temporal lobe            26312
#> c43ac461-9f03-44bc-be7d-3d867eb708a0             Stage III 2019-03-14T10:37:34.405260-05:00       Floor of mouth, NOS            25635
#> a59a90d9-f1b0-49dd-9c97-bcaa6ba55d44          Not Reported 2019-03-14T10:37:34.405260-05:00       Floor of mouth, NOS            16652
#> 59122a43-606a-4669-806b-6747e0ac9985          Not Reported 2019-07-22T06:40:02.183501-05:00          Upper lobe, lung            23384
#> 4447a969-e5c8-4291-b83c-53a0f7e77cbc          Not Reported 2019-05-07T07:41:33.411909-05:00              Frontal lobe            29326
```

# Basic design

This package design is meant to have some similarities to the
“tidyverse” approach of dplyr. Roughly, the functionality for finding
and accessing files and metadata can be divided into:

1.  Simple query constructors based on GDC API endpoints.
2.  A set of verbs that when applied, adjust filtering, field selection,
    and faceting (fields for aggregation) and result in a new query
    object (an endomorphism)
3.  A set of verbs that take a query and return results from the GDC

In addition, there are auxiliary functions for asking the GDC API for
information about available and default fields, slicing BAM files, and
downloading actual data files. Here is an overview of functionality[^1].

-   Creating a query
    -   `projects()`
    -   `cases()`
    -   `files()`
    -   `annotations()`
-   Manipulating a query
    -   `filter()`
    -   `facet()`
    -   `select()`
-   Introspection on the GDC API fields
    -   `mapping()`
    -   `available_fields()`
    -   `default_fields()`
    -   `grep_fields()`
    -   `available_values()`
    -   `available_expand()`
-   Executing an API call to retrieve query results
    -   `results()`
    -   `count()`
    -   `response()`
-   Raw data file downloads
    -   `gdcdata()`
    -   `transfer()`
    -   `gdc_client()`
-   Summarizing and aggregating field values (faceting)
    -   `aggregations()`
-   Authentication
    -   `gdc_token()`
-   BAM file slicing
    -   `slicing()`

[^1]: See individual function and methods documentation for specific
    details.

---
output: github_document
knit: (function(inputFile, encoding) {
  rmarkdown::render(inputFile, encoding = encoding, output_dir = "../../") })
---

```{r, include = FALSE}
knitr::opts_chunk$set(
  collapse = TRUE,
  comment = "#>",
  cache = TRUE,
  out.width = "100%"
)
```

```{r,echo=FALSE,include=FALSE,eval=FALSE}
rmarkdown::render("inst/script/README.Rmd", output_dir = ".")
```


# GenomicDataCommons

<!-- badges: start -->
[![R-CMD-check](https://github.com/Bioconductor/GenomicDataCommons/workflows/R-CMD-check/badge.svg)](https://github.com/Bioconductor/GenomicDataCommons/actions)
<!-- badges: end -->

# What is the GDC?

From the [Genomic Data Commons (GDC) website](https://gdc.nci.nih.gov/about-gdc):

The National Cancer Institute's (NCI's) Genomic Data Commons (GDC) is
a data sharing platform that promotes precision medicine in
oncology. It is not just a database or a tool; it is an expandable
knowledge network supporting the import and standardization of genomic
and clinical data from cancer research programs.

The GDC contains NCI-generated data from some of the largest and most
comprehensive cancer genomic datasets, including The Cancer Genome
Atlas (TCGA) and Therapeutically Applicable Research to Generate
Effective Therapies (TARGET). For the first time, these datasets have
been harmonized using a common set of bioinformatics pipelines, so
that the data can be directly compared.

As a growing knowledge system for cancer, the GDC also enables
researchers to submit data, and harmonizes these data for import into
the GDC. As more researchers add clinical and genomic data to the GDC,
it will become an even more powerful tool for making discoveries about
the molecular basis of cancer that may lead to better care for
patients.

The
[data model for the GDC is complex](https://gdc.cancer.gov/developers/gdc-data-model/gdc-data-model-components),
but it worth a quick overview. The data model is encoded as a
so-called property graph. Nodes represent entities such as Projects,
Cases, Diagnoses, Files (various kinds), and Annotations. The
relationships between these entities are maintained as edges.  Both
nodes and edges may have Properties that supply instance details.  The
GDC API exposes these nodes and edges in a somewhat simplified set
of
[RESTful](https://en.wikipedia.org/wiki/Representational_state_transfer)
endpoints.

# Quickstart

This software is available at Bioconductor.org and can be downloaded via
`BiocManager::install`.

To report bugs or problems, either
[submit a new issue](https://github.com/Bioconductor/GenomicDataCommons/issues)
or submit a `bug.report(package='GenomicDataCommons')` from within R (which
will redirect you to the new issue on GitHub).

## Installation

Installation can be achieved via Bioconductor's `BiocManager` package.

```{r,eval=FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")

BiocManager::install('GenomicDataCommons')
```

```{r,include=TRUE,results="hide",message=FALSE,warning=FALSE}
library(GenomicDataCommons)
```

## Check basic functionality

```{r}
status()
```

## Find data

The following code builds a `manifest` that can be used to guide the
download of raw data. Here, filtering finds gene expression files
quantified as raw counts using `STAR` from ovarian cancer patients.

```{r}
ge_manifest <- files() |>
    filter( cases.project.project_id == 'TCGA-OV') |>
    filter( type == 'gene_expression' ) |>
    filter( analysis.workflow_type == 'STAR - Counts') |>
    manifest(size = 5)
ge_manifest
```

## Download data

This code block downloads the `r nrow(ge_manifest)` gene expression files
specified in the query above. Using multiple processes to do the download very
significantly speeds up the transfer in many cases.  The following completes in
about 15 seconds.

```{r,eval=FALSE}
library(BiocParallel)
register(MulticoreParam())
destdir <- tempdir()
fnames <- lapply(ge_manifest$id,gdcdata)
```

If the download had included controlled-access data, the download above would
have needed to include a `token`.  Details are available in
[the authentication section below](#authentication).

## Metadata queries

Here we use a couple of ad-hoc helper functions to handle the output of the
query. See the `inst/script/README.Rmd` folder for the source.

```{r,echo=FALSE}
filterAllNA <- function(df) {
    notallna <- vapply(df, function(x) !all(is.na(x)), logical(1L))
    df[, notallna]
}

bindrowname <- function(resultList) {
    if (is.data.frame(resultList))
        stop("Only run this on the list type of outputs")
    datadf <- dplyr::bind_rows(resultList)
    rownames(datadf) <- names(resultList)
    filterAllNA(datadf)
}
```

First, create a `data.frame` from the clinical data:

```{r}
expands <- c("diagnoses","annotations",
             "demographic","exposures")
clinResults <- cases() |>
    GenomicDataCommons::select(NULL) |>
    GenomicDataCommons::expand(expands) |>
    results(size=6)
demoDF <- filterAllNA(clinResults$demographic)
exposuresDF <- bindrowname(clinResults$exposures)
```

```{r}
demoDF[, 1:4]
```

```{r}
exposuresDF[, 1:4]
```

Note that the diagnoses data has multiple lines per patient:

```{r}
diagDF <- bindrowname(clinResults$diagnoses)
diagDF[, 1:4]
```

# Basic design

This package design is meant to have some similarities to the "tidyverse"
approach of dplyr. Roughly, the functionality for finding and accessing files
and metadata can be divided into:

1. Simple query constructors based on GDC API endpoints.
2. A set of verbs that when applied, adjust filtering, field selection, and
faceting (fields for aggregation) and result in a new query object (an
endomorphism)
3. A set of verbs that take a query and return results from the GDC

In addition, there are auxiliary functions for asking the GDC API for
information about available and default fields, slicing BAM files, and
downloading actual data files.  Here is an overview of functionality[^1].


- Creating a query
    - `projects()`
    - `cases()`
    - `files()`
    - `annotations()`
- Manipulating a query
    - `filter()`
    - `facet()`
    - `select()`
- Introspection on the GDC API fields
    - `mapping()`
    - `available_fields()`
    - `default_fields()`
    - `grep_fields()`
    - `available_values()`
    - `available_expand()`
- Executing an API call to retrieve query results
    - `results()`
    - `count()`
    - `response()`
- Raw data file downloads
    - `gdcdata()`
    - `transfer()`
    - `gdc_client()`
- Summarizing and aggregating field values (faceting)
    - `aggregations()`
- Authentication
    - `gdc_token()`
- BAM file slicing
    - `slicing()`

[^1]: See individual function and methods documentation for specific details.

## Running tests

```{r}
devtools::test()
```

Should also run under `R CMD BiocCheck/check`.

## Tests

A test file lives in tests/testthat/. Its name must start with test. Here’s an example of a test file from the stringr package:

```{r}
library(stringr)
context("String length")

test_that("str_length is number of characters", {
  expect_equal(str_length("a"), 1)
  expect_equal(str_length("ab"), 2)
  expect_equal(str_length("abc"), 3)
})

test_that("str_length of factor is length of level", {
  expect_equal(str_length(factor("a")), 1)
  expect_equal(str_length(factor("ab")), 2)
  expect_equal(str_length(factor("abc")), 3)
})

test_that("str_length of missing is missing", {
  expect_equal(str_length(NA), NA_integer_)
  expect_equal(str_length(c(NA, 1)), c(NA, 1))
  expect_equal(str_length("NA"), 2)
})
```

Tests are organised hierarchically: expectations are grouped into tests which are organised in files:

An expectation is the atom of testing. It describes the expected result of a computation: Does it have the right value and right class? Does it produce error messages when it should? An expectation automates visual checking of results in the console. Expectations are functions that start with expect_.

A test groups together multiple expectations to test the output from a simple function, a range of possibilities for a single parameter from a more complicated function, or tightly related functionality from across multiple functions. This is why they are sometimes called unit as they test one unit of functionality. A test is created with test_that() .

A file groups together multiple related tests. Files are given a human readable name with context().

#' Query the GDC for current status
#'
#' @param version (optional) character(1) version of GDC
#'
#' @return List describing current status.
#' 
#' @importFrom httr content
#'
#' @examples
#' status()
#' 
#' @export
status <- function(version=NULL) {
    response <- .gdc_get(paste(version, "status", sep="/"),archive='default')
    content(response, type="application/json")
}

#' Get clinical information from GDC
#'
#' The NCI GDC has a complex data model that allows various studies to
#' supply numerous clinical and demographic data elements. However,
#' across all projects that enter the GDC, there are
#' similarities. This function returns four data.frames associated
#' with case_ids from the GDC.
#'
#' @param case_ids a character() vector of case_ids, typically from
#'     "cases" query.
#'
#' @param include_list_cols logical(1), whether to include list
#'     columns in the "main" data.frame. These list columns have
#'     values for aliquots, samples, etc. While these may be useful
#'     for some situations, they are generally not that useful as
#'     clinical annotations.
#' 
#' @importFrom jsonlite fromJSON
#' @importFrom dplyr bind_rows
#' @importFrom tibble as_tibble
#'
#' @details
#' Note that these data.frames can, in general, have different numbers
#' of rows (or even no rows at all). If one wishes to combine to
#' produce a single data.frame, using the approach of left joining to
#' the "main" data.frame will yield a useful combined data.frame. We
#' do not do that directly given the potential for 1:many
#' relationships. It is up to the user to determine what the best
#' approach is for any given dataset.
#'
#'
#' @return
#' A list of four data.frames:
#' \enumerate{
#' \item main, representing basic case identification and metadata
#'     (update date, etc.)
#' \item diagnoses
#' \item esposures
#' \item demographic
#' }
#'
#'
#' @examples
#' case_ids = cases() |> results(size=10) |> ids()
#' clinical_data = gdc_clinical(case_ids)
#'
#' # overview of clinical results
#' class(clinical_data)
#' names(clinical_data)
#' sapply(clinical_data, class)
#' sapply(clinical_data, nrow)
#'
#' # available data
#' head(clinical_data$main)
#' head(clinical_data$demographic)
#' head(clinical_data$diagnoses)
#' head(clinical_data$exposures)
#' 
#' @export
gdc_clinical = function(case_ids, include_list_cols = FALSE) {
    stopifnot(is.character(case_ids))
    stopifnot(is.logical(include_list_cols) & length(include_list_cols)==1)
    resp = cases() |>
        filter( ~ case_id %in% case_ids) |>
        expand(c("diagnoses",
                 "demographic",
                 "exposures",
                 "follow_ups.other_clinical_attributes")) |>
        response_all(response_handler = function(x) jsonlite::fromJSON(x, simplifyDataFrame = TRUE))
    demographic = resp$results$demographic
    demographic$case_id = rownames(demographic)

    nodx <- vapply(resp$results$diagnoses, is.null, logical(1L))
    if (any(nodx))
        resp$results$diagnoses[nodx] <- list(data.frame())
    
    diagnoses <- suppressMessages({
        bind_rows(
            lapply(resp$results$diagnoses, readr::type_convert),
            .id = "case_id"
        )
    })

    exposures = bind_rows(resp$results$exposures, .id = "case_id")
    follow_ups = bind_rows(resp$results$follow_ups, .id = "case_id")

    # set up main table by removing data.frame columns
    cnames = setdiff(colnames(resp$results), c('exposures', 'follow_ups', 'diagnoses', 'demographic'))
    main = resp$results[, cnames]

    if(!include_list_cols) {
        non_list_cols = names(Filter(function(cname) cname!='list', sapply(main, class)))
        main = main[, non_list_cols]
    }
    
    y = list(demographic = as_tibble(demographic),
             diagnoses = as_tibble(diagnoses),
             exposures = as_tibble(exposures),
             follow_ups = as_tibble(follow_ups),
             main = as_tibble(main))
    class(y) = c('GDCClinicalList', class(y))
    return(y)
}

test_that("clinical data is structured properly", {
    sizen <- 3
    case_ids <- cases() |> results(size=sizen) |> ids()
    clinical_data <- gdc_clinical(case_ids)
    # overview of clinical results
    expect_true(
        is(clinical_data, "GDCClinicalList")
    )
    expect_true(
        all(
            c("demographic", "diagnoses", "exposures", "follow_ups", "main")
            %in%
            names(clinical_data)
        )
    )
    
    ## exposures has no rows
    clinical_data <- clinical_data[names(clinical_data) != "exposures"]
    expect_true(
        all(
            vapply(clinical_data, nrow, integer(1L)) >= sizen
        )
    )
    expect_true(
        all(
            vapply(clinical_data, is.data.frame, logical(1L))
        )
    )
})

#' Read a single htseq-counts result file.
#'
#' The htseq package is used extensively to count reads
#' relative to regions (see 
#' \url{http://www-huber.embl.de/HTSeq/doc/counting.html}).
#' The output of htseq-count is a simple two-column table
#' that includes features in column 1 and counts in column 2.
#' This function simply reads in the data from one such file
#' and assigns column names. 
#'
#' @param fname character(1), the path of the htseq-count file.
#' @param samplename character(1), the name of the sample. This will
#'     become the name of the second column on the resulting
#'     \code{data.frame}, making for easier merging if necessary.
#' @param ... passed to \code{\link[readr]{read_tsv})}
#' @return a two-column data frame
#'
#' @examples
#' fname = system.file(package='GenomicDataCommons',
#'                     'extdata/example.htseq.counts.gz')
#' dat = readHTSeqFile(fname)
#' head(dat)
#'
#' @export
readHTSeqFile <- function(fname, samplename = 'sample', ...) {
    if(!file.exists(fname))
        stop(sprintf('The specified file, %s, does not exist',fname))
    if(!((length(fname) == 1) & (is.character(fname))))
        stop('fname must be of type character(1)')
    tmp = read_tsv(fname,col_names = FALSE)
    if(ncol(tmp) != 2)
        stop(sprintf('%s had %d columns, expected 2 columns',fname, ncol(tmp)))
    colnames(tmp) = c('feature',samplename)
    tmp
}

library(GenomicDataCommons)
context('readHTSeqFile')

test_that("readHTSeqFile works on example data", {
    dat = readHTSeqFile(system.file(package="GenomicDataCommons",
                                    'extdata/example.htseq.counts.gz'))
    expect_equal(nrow(dat),50)
    expect_equal(ncol(dat),2)
})

#' Work with gdc cache directory
#'
#' The GenomicDataCommons package will cache downloaded
#' files to minimize network and allow for
#' offline work. These functions are used to create a cache directory
#' if one does not exist, set a global option, and query that
#' option. The cache directory will default to the user "cache"
#' directory according to specifications in
#' \code{\link[rappdirs]{app_dir}}. However, the user may want to set
#' this to another direcotory with more or higher performance
#' storage. 
#'
#' @return character(1) directory path that serves as
#' the base directory for GenomicDataCommons downloads.
#'
#' @details
#' The cache structure is currently just a directory with each file
#'     being represented by a path constructed as:
#'     CACHEDIR/UUID/FILENAME. The cached files can be manipulated
#'     using standard file system commands (removing, finding,
#'     etc.). In this sense, the cache sytem is minimalist in design.
#' 
#' @examples
#' gdc_cache()
#' \dontrun{
#' gdc_set_cache(getwd())
#' }
#' 
#' @export
gdc_cache = function()
{
    cache_dir = getOption('gdc_cache',gdc_set_cache(verbose=FALSE))
    if(!dir.exists(cache_dir)) {
        gdc_set_cache(cache_dir)
    }
    return(cache_dir)
}

#' @describeIn gdc_cache (Re)set the GenomicDataCommons cache
#'     directory
#' 
#' @importFrom utils menu
#' @importFrom rappdirs app_dir
#'
#'
#' @param create_without_asking logical(1) specifying whether to allow
#'     the function to create the cache directory without asking the
#'     user first. In an interactive session, if the cache directory
#'     does not exist, the user will be prompted before creation.
#'
#' @param verbose logical(1) whether or not to message the location of
#'     the cache directory after creation.
#'
#' @param directory character(1) directory path, will be created
#'     recursively if not present.
#'
#'
#' @return the created directory (invisibly)
#' 
#' @export
gdc_set_cache = function(directory = rappdirs::app_dir(appname =
                                                           "GenomicDataCommons")$cache(),
                         verbose = TRUE,
                         create_without_asking = !interactive())
{

    create_path = function(directory) {
        dir.create(directory, recursive = TRUE, showWarnings =
                                                    FALSE)
    }
    
    if(is.character(directory) & length(directory)==1) {
        # if directory exists, move on
        if(!dir.exists(directory)) {
            # if not in an interactive session, go
            # ahead and create directory without user
            # input.
            if(create_without_asking) {
                create_path(directory)
            } else {
                # If in an interactive environment,
                # go ahead and ask user for agreement.
                response = menu(c("Yes", "No"),
                                title=sprintf("Would you like to create a GDC Cache directory at %s", directory))
                if(response == 1) {
                    create_path(directory)
                } else {
                    stop("GDC Cache directory cannot be created without user agreement")
                }
            }
        }
        options('gdc_cache' = directory)
    } else {
        stop("directory should be a character(1)")
    }
    if(verbose) message("GDC Cache directory set to: ", directory)
    invisible(directory)
}

library(GenomicDataCommons)
context('cache_control')

cache = gdc_cache()

test_that("getting cache returns length 1 char vector", {
    expect_length(gdc_cache(),1)
    expect_true(is.character(gdc_cache()))
})

test_that("setting cache works", {
    expect_equal(gdc_set_cache('/tmp'),'/tmp')
    expect_equal(gdc_cache(),'/tmp')
})

test_that("setting cache error checking works", {
    expect_error(gdc_set_cache(1))
    expect_error(gdc_set_cache(c('a','b')))
})

gdc_set_cache(cache)

#' GenomicDataCommons: A package for interfacing with the NCI GDC
#' 
#' @section finding data:
#' 
#' \itemize{
#' \item{\code{\link{query}}}
#' \item{\code{\link{cases}}}
#' \item{\code{\link{projects}}}
#' \item{\code{\link{files}}}
#' \item{\code{\link{annotations}}}
#' \item{\code{\link{mapping}}}
#' }
#'
#' @section downloading data:
#' data
#'
"_PACKAGE"

#' S3 Generic to return all GDC fields
#'
#' @param x A character(1) string ('cases','files','projects',
#' 'annotations') or an subclass of \code{\link{GDCQuery}}.
#' @return a character vector of the default fields
#'
#' @examples
#' available_fields('projects')
#' projQuery = query('projects')
#' available_fields(projQuery)
#' 
#' @export
available_fields = function(x) {
    UseMethod('available_fields',x)
}

#' @describeIn available_fields GDCQuery method
#' @export
available_fields.GDCQuery = function(x) {
    return(mapping(entity_name(x))$field)
}

#' @describeIn available_fields character method
#' @export
available_fields.character = function(x) {
    stopifnot(length(x)==1,x %in% .gdc_entities)
    return(mapping(x)$field)
}


#' S3 Generic to return default GDC fields
#'
#' @param x A character string ('cases','files','projects',
#' 'annotations') or an subclass of \code{\link{GDCQuery}}.
#' @return a character vector of the default fields
#'
#' @examples
#' default_fields('projects')
#' projQuery = query('projects')
#' default_fields(projQuery)
#' 
#' @export
default_fields = function(x) {
    UseMethod('default_fields',x)
}

#' @describeIn default_fields character method
#' @export
default_fields.character = function(x) {
    defaults=NA # just to avoid no visible binding note
    stopifnot(length(x)==1,x %in% .gdc_entities)
    return(subset(mapping(x),defaults)$field)
}

#' @describeIn default_fields GDCQuery method
#' @export
default_fields.GDCQuery = function(x) {
    return(default_fields(entity_name(x)))
}

#' S3 generic to set GDCQuery fields
#'
#' @param x the objects on which to set fields
#' @param fields a character vector specifying the fields
#' 
#'
#' @return A \code{\link{GDCQuery}} object, with the fields
#' member altered.
#' 
#' @examples
#' gProj = projects()
#' gProj$fields
#' head(available_fields(gProj))
#' default_fields(gProj)
#'
#' gProj |>
#'   select(default_fields(gProj)[1:2]) |>
#'   response() |>
#'   str(max_level=2)
#' 
#' @export
select <- function(x,fields) {
    UseMethod('select',x)
}



#" (internal) rectify specified fields with available fields
.gdcRectifyFieldsForEntity <- function(entity,fields) {
    af = available_fields(entity)
    mismatches = fields[!(fields %in% af)]
    if(length(mismatches)>0)
        stop(sprintf('fields specified included fields not available in %s including (%s)',entity,mismatches))
    fields = union(paste0(sub('s$','',entity),"_id"),fields)
    return(fields)
}

#' @describeIn select set fields on a GDCQuery object
#' @export
select.GDCQuery <- function(x,fields) {
    x$fields = .gdcRectifyFieldsForEntity(entity_name(x),fields)
    return(x)
}

#' Find matching field names
#' 
#' This utility function allows quick text-based search of available
#' fields for using \code{\link{grep}}
#' 
#' @param entity one of the available gdc entities ('files','cases',...)
#'     against which to gather available fields for matching
#' 
#' @param pattern A regular expression that will be used
#' in a call to \code{\link{grep}}
#' 
#' @param ... passed on to grep
#' 
#' @param value logical(1) whether to return values as opposed
#' to indices (passed along to grep)
#'
#' @return character() vector of field names matching
#'     \code{pattern}
#' 
#' @examples 
#' grep_fields('files','analysis')
#' 
#' @export
grep_fields <- function(entity,pattern,...,value=TRUE) {
  stopifnot(entity %in% .gdc_entities)
  return(grep(pattern=pattern,
              x=available_fields(entity),
              value=TRUE,...))
}

#' Find common values for a GDC field
#' 
#' @param entity character(1), a GDC entity ("cases", "files", "annotations", "projects")
#' @param field character(1), a field that is present in the entity record
#'
#' @return character vector of the top 100 (or fewer) most frequent
#'     values for a the given field
#' 
#' @examples 
#' available_values('files','cases.project.project_id')[1:5]
#' 
#' @export
available_values <- function(entity,field) {
    stopifnot(entity %in% .gdc_entities)
    agg = query(entity) |> facet(field) |> aggregations()
    agg[[field]]$key
}

#' S3 Generic that returns the field description text, if available
#'
#' @param entity character(1) string ('cases','files','projects',
#' 'annotations', etc.) or an subclass of \code{\link{GDCQuery}}.
#'
#' @param field character(1), the name of the field that will be used to look
#' up the description.
#' 
#' @return character(1) descriptive text or character(0) if no description
#' is available.
#'
#' @examples
#' field_description('cases', 'annotations.category')
#' casesQuery = query('cases')
#' field_description(casesQuery, 'annotations.category')
#' field_description(cases(), 'annotations.category')
#' 
#' @export
field_description = function(entity, field) {
    UseMethod('field_description',entity)
}

#' @describeIn field_description GDCQuery method
#' @export
field_description.GDCQuery = function(entity, field) {
    stopifnot(length(field)==1)
    m = mapping(entity_name(entity))
    return(m$description[m$field==field])
}

#' @describeIn field_description character method
#' @export
field_description.character = function(entity, field) {
    stopifnot(length(entity)==1,entity %in% .gdc_entities)
    stopifnot(length(field)==1)
    m = mapping(entity)
    return(m$description[m$field==field])
}

#' Bulk data download
#'
#' The GDC maintains a special tool,
#' \href{the GDC Data Transfer Tool}{https://docs.gdc.cancer.gov/Data_Transfer_Tool/Users_Guide/Getting_Started/},
#' that enables high-performance, potentially parallel, and
#' resumable downloads. The Data Transfer Tool is an external
#' program that requires separate download. Due to recent changes in the
#' GDC API, the transfer function now validates the version of the `gdc-client`
#' to ensure reliable downloads.
#'
#' @param uuids character() vector of GDC file UUIDs
#'
#' @param args character() vector specifying command-line arguments to
#'     be passed to \code{gdc-client}. See \code{\link{transfer_help}} for
#'     possible values. The arguments \code{--manifest}, \code{--dir},
#'     and \code{--token-file} are determined by \code{manifest},
#'     \code{destination_dir}, and \code{token}, respectively, and
#'     should NOT be provided as elements of \code{args}.
#'
#' @param token character(1) containing security
#'     token allowing access to restricted data. See
#'     \url{https://gdc-docs.nci.nih.gov/API/Users_Guide/Authentication_and_Authorization/}.
#'     Note that the GDC transfer tool requires a file for data
#'     transfer. Therefore, this token will be written to a temporary
#'     file (with appropriate permissions set).
#'
#' @param overwrite logical(1) default FALSE indicating whether
#'     existing files with identical name should be over-written.
#'
#' @return character(1) directory path to which the files were
#'     downloaded.
#'
#' @examples
#' \dontrun{
#' uuids = files() |> 
#'   filter(access == "open") |> 
#'   results() |>
#'   ids()
#' file_paths <- transfer(uuids)
#' file_paths
#' names(file_paths)
#' # and with authenication
#' # REQUIRES gdc_token 
#' # destination <- transfer(uuids,token=gdc_token())
#' }
#'
#' @importFrom utils read.table
#' @export
transfer <-
    function(uuids, args=character(), token=NULL, overwrite=FALSE)
    {
        stopifnot(is.character(uuids))
        destination_dir <- gdc_cache()

        manifest = files() |>
            GenomicDataCommons::filter( ~ file_id %in% uuids ) |>
            GenomicDataCommons::manifest()
        manifest_file = write_manifest(manifest)
        

        dir_arg <- sprintf("--dir %s", destination_dir)
        manifest_arg <- sprintf("--manifest %s", manifest_file)
        token_file = tempfile()
        if (!is.null(token)) {
            writeLines(token,con=token_file)
            stopifnot(file.exists(token_file))
            Sys.chmod(token_file,mode="600")
            token <- sprintf("--token-file %s", token_file)
        }
        gdc_client_version_validate()
        args <- paste(c("download", dir_arg, manifest_arg, args, token), collapse=" ")
        system2(gdc_client(), args)

        if(!is.null(token))
            unlink(token_file)

        filepaths <- file.path(gdc_cache(), uuids,
                               as.character(manifest[[2]]))
        names(filepaths) = uuids
        return(filepaths)
    }

#' return gdc-client executable path
#' 
#' This function is a convenience function to 
#' find and return the path to the GDC Data Transfer
#' Tool executable assumed to be named 'gdc-client'. 
#' The assumption is that the appropriate version of the
#' GDC Data Transfer Tool is a separate download available
#' from \href{the GDC website}{https://gdc.cancer.gov/access-data/gdc-data-transfer-tool}
#' and as a backup from \href{on github}{https://github.com/NCI-GDC/gdc-client}.
#'
#' @details
#' The path is checked in the following order:
#' \enumerate{
#' \item an R option("gdc_client")
#' \item an environment variable GDC_CLIENT
#' \item from the search PATH
#' \item in the current working directory
#' }
#'
#' @return character(1) the path to the gdc-client executable.
#'
#' @examples
#' # this cannot run without first
#' # downloading the GDC Data Transfer Tool
#' gdc_client = try(gdc_client(),silent=TRUE)
#' 
#' @export
gdc_client = function() {
    if(!is.null(getOption('gdc_client')))
        if(file.exists(getOption('gdc_client')))
            return(getOption('gdc_client'))
    if(file.exists(Sys.getenv("GDC_CLIENT")))
        return(Sys.getenv("GDC_CLIENT"))
    if(!(Sys.which("gdc-client")==''))
        return(Sys.which("gdc-client"))
    client=dir('.',pattern='^gdc-client$',full.names=TRUE)
    if(length(client)==1)
        if(client=='./gdc-client')
            return(client)
    stop('gdc_client not found. Be sure to install the command \nline GDC client available from the GDC website.')
}

gdc_client_version <- function() {
    gc_loc <- gdc_client()
    vers <- system2(gc_loc, "--version", stdout = TRUE, stderr = TRUE)
    vers <- gsub("^v", "", vers)
    package_version(vers)
}

.GDC_COMPATIBLE_VERSION <- "1.3.0"

#' @describeIn transfer
#'
#' If you are using the 'client' option, your `gdc-client` should be
#' up-to-date (>= 1.3.0).
#'
#' @param valid_version character(1) The last known version that works for the
#'     current data release for which to validate against, not typically changed
#'     by the end-user.
#'
#' @export
gdc_client_version_validate <-
    function(valid_version = .GDC_COMPATIBLE_VERSION)
{
    client_ver <- gdc_client_version()
    if (client_ver < package_version(valid_version))
        stop("Update the 'gdc_client' to a version >= ", valid_version)
}

#' \code{transfer_help()} queries the the command line GDC Data
#' Transfer Tool, \code{gdc-client}, for available options to be used
#' in the \code{\link{transfer}} command.
#'
#' @describeIn transfer
#'
#' @export
transfer_help <- function() {
    system2(gdc_client(), "download -h")
}

#' Get the entity name from a GDCQuery object
#'
#' An "entity" is simply one of the four medata endpoints.
#' \itemize{
#' \item{cases}
#' \item{projects}
#' \item{files}
#' \item{annotations}
#' }
#' All \code{\link{GDCQuery}} objects will have an entity name. This S3 method
#' is simply a utility accessor for those names.
#' 
#' @param x a \code{\link{GDCQuery}} object
#'
#' @return character(1) name of an associated entity; one of
#' "cases", "files", "projects", "annotations".
#' 
#' @examples
#' qcases = cases()
#' qprojects = projects()
#' 
#' entity_name(qcases)
#' entity_name(qprojects)
#' 
#' @export
entity_name = function(x) {
    UseMethod('entity_name',x)
}


#' @rdname entity_name
#' @export
entity_name.GDCQuery = function(x) {
    cls = class(x)[1]
    return(substr(cls,5,nchar(cls)))
}

#' @rdname entity_name
#' @export
entity_name.GDCResults = function(x) {
    cls = class(x)[1]
    return(substr(cls,4,nchar(cls)-8))
}

#' Fetch \code{\link{GDCQuery}} metadata from GDC
#'
#' @aliases GDCResponse
#' 
#' @param x a \code{\link{GDCQuery}} object
#' @param from integer index from which to start returning data
#' @param size number of records to return
#' @param ... passed to httr (good for passing config info, etc.)
#' @param response_handler a function that processes JSON (as text)
#' and returns an R object.  Default is \code{\link[jsonlite]{fromJSON}}.
#' 
#' @rdname response
#'
#' @return A \code{GDCResponse} object which is a list with the following
#' members:
#' \itemize{
#' \item{results}
#' \item{query}
#' \item{aggregations}
#' \item{pages}
#' }
#' 
#' 
#' @examples
#'
#' # basic class stuff
#' gCases = cases()
#' resp = response(gCases)
#' class(resp)
#' names(resp)
#'
#' # And results from query
#' resp$results[[1]]
#' 
#' @export
response = function(x,...) {
    UseMethod('response',x)
}

#' provide count of records in a \code{\link{GDCQuery}}
#'
#' @param x a \code{\link{GDCQuery}} object
#' @param ... passed to httr (good for passing config info, etc.)
#'
#' @return integer(1) representing the count of records that will
#'  be returned by the current query
#' 
#' @examples
#' # total number of projects
#' projects() |> count()
#'
#' # total number of cases
#' cases() |> count()
#' 
#' @export
count = function(x,...) {
    UseMethod('count',x)
}

#' @describeIn count
#'
#' @export
count.GDCQuery = function(x,...) {
    resp = x |> response(size=1)
    return(resp$pages$total)
}    

#' @describeIn count
#'
#' @export
count.GDCResponse = function(x,...) {
    x$pages$total
}


#" (internal) prepare "results" for return
#"
#" In particular, this function sets
#" entity_ids for every element so that
#" one does not loose track of the relationships
#" given the nested nature of GDC returns
.prepareResults <- function(res,idfield) {
    for(i in names(res)) {
        if(inherits(res[[i]],'data.frame'))
            rownames(res[[i]]) = res[[idfield]]
        else
            names(res[[i]]) = res[[idfield]]}
    return(res)
}

#' @rdname response
#' 
#' @importFrom jsonlite fromJSON
#' 
#' @export
response.GDCQuery = function(x, from = 0, size = 10, ...,
                             response_handler = jsonlite::fromJSON) {
    body = Filter(function(z) !is.null(z),x)
    body[['facets']]=paste0(body[['facets']],collapse=",")
    body[['fields']]=paste0(body[['fields']],collapse=",")
    body[['expand']]=paste0(body[['expand']],collapse=",")
    body[['from']]=from
    body[['size']]=size
    body[['format']]='JSON'
    body[['pretty']]='FALSE'
    tmp = response_handler(httr::content(
      .gdc_post(entity_name(x),body=body, token=NULL,...),
                                         as="text", encoding = "UTF-8"))
    res = tmp$data$hits
    idfield = paste0(sub('s$','',entity_name(x)),'_id')
    ## the following code just sets names on the 
    structure(
        list(results = .prepareResults(res,idfield),
             query   = x,
             pages   = tmp$data$pagination,
             aggregations = lapply(tmp$data$aggregations,function(x) {x$buckets})),
        class = c(paste0('GDC',entity_name(x),'Response'),'GDCResponse','list')
    )
}

#' @rdname response
#' 
#' @export
response_all = function(x,...) {
    count = count(x)
    return(response(x=x,size=count,from=0,...))
}


#' aggregations
#'
#' @param x a \code{\link{GDCQuery}} object
#'
#' @return a \code{list} of \code{data.frame} with one
#' member for each requested facet. The data frames
#' each have two columns, key and doc_count.
#' 
#' @examples
#' # Number of each file type
#' res = files() |> facet(c('type','data_type')) |> aggregations()
#' res$type
#'
#' @export
aggregations = function(x) {
    UseMethod('aggregations',x)
}


#' @describeIn aggregations
#'
#'
#' @export
aggregations.GDCQuery = function(x) {
    if(is.null(x$facets))
        x = x |> facet()
    return(response(x)$aggregations)
}

#' @describeIn aggregations
#'
#'
#' @export
aggregations.GDCResponse = function(x) {
    x$aggregations
}


#' results
#'
#' @param x a \code{\link{GDCQuery}} object
#' @param ... passed on to \code{\link{response}}
#' 
#' @return A (typically nested) \code{list} of GDC records
#' 
#' @examples
#' qcases = cases() |> results()
#' length(qcases)
#'
#' @export
results = function(x,...) {
    UseMethod('results',x)
}

#' results_all
#'
#' @param x a \code{\link{GDCQuery}} object
#'
#' @return A (typically nested) \code{list} of GDC records
#' 
#' @examples
#' # details of all available projects
#' projResults = projects() |> results_all()
#' length(projResults)
#' count(projects())
#'
#' 
#' @export
results_all = function(x) {
    UseMethod('results_all',x)
}


#' @describeIn results
#'
#'
#' @export
results.GDCQuery = function(x,...) {
    results(response(x,...))
}

#' @describeIn results_all
#'
#'
#' @export
results_all.GDCQuery = function(x) {
    results(response_all(x))
}

#' @describeIn results
#'
#'
#' @export
results.GDCResponse = function(x,...) {
    structure(
        x$results,
        class=c(sub('Response','Results',class(x)))
    )
}

#' @describeIn results_all
#'
#'
#' @export
results_all.GDCResponse = function(x) {
    structure(
        x$results,
        class=c(sub('Response','Results',class(x)))
    )
}




#' @importFrom xml2 xml_find_all
.response_warnings <- function(warnings, endpoint)
{
    warnings <- vapply(warnings, as.character, character(1))
    if (length(warnings) && nzchar(warnings))
        warning("'", endpoint, "' query warnings:\n", .wrapstr(warnings))
    NULL
}

.response_json_as_list <- function(json, endpoint)
{
    type <- substr(endpoint, 1, nchar(endpoint) - 1L)
    type_id <- sprintf("%s_id", type)
    type_list <- sprintf("%ss_list", type)

    hits <- json[["data"]][["hits"]]
    names(hits) <- vapply(hits, "[[", character(1), type_id)
    hits <- lapply(hits, "[[<-", type_id, NULL)
    hits <- lapply(hits, lapply, unlist) # collapse field elt 'list'
    class(hits) <- c(type_list, "gdc_list", "list")
    hits
}   

#' @importFrom stats setNames
#' @importFrom xml2 xml_find_all xml_text
.response_xml_as_data_frame <- function(xml, fields)
{
    xpaths <- setNames(sprintf("/response/data/hits/item/%s", fields), fields)

    columns <- lapply(xpaths, function(xpath, xml) {
        nodes <- xml_find_all(xml, xpath)
        vapply(nodes, xml_text, character(1))
    }, xml=xml)
    columns <- Filter(length, columns)

    dropped <- fields[!fields %in% names(columns)]
    if (length(dropped))
        warning("fields not available:\n", .wrapstr(dropped))
    if (length(columns)==0) {
      warning("No records found. Check on filter criteria to ensure they do what you expect. ")
      return(NULL)
    }
    if (!length(unique(lengths(columns)))) {
        lens <- paste(sprintf("%s = %d", names(columns), lengths(columns)),
                      collapse=", ")
        stop("fields are different lengths:\n", .wrapstr(lens))
    }

    as.data.frame(columns, stringsAsFactors=FALSE)
}

#' return a gdc token from file or environment
#' 
#' The GDC requires an auth token for downloading
#' data that are "controlled access". For example, 
#' BAM files for human datasets, germline variant calls,
#' and SNP array raw data all are protected as "controlled
#' access". For these files, a GDC access token is required.
#' See the \href{details on the GDC authentication and token information}{https://docs.gdc.cancer.gov/Data_Portal/Users_Guide/Authentication/#gdc-authentication-tokens}.
#' Note that this function simply returns a string value. 
#' It is possible to keep the GDC token in a variable in R
#' or to pass a string directly to the appropriate parameter.
#' This function is simply a convenience function for alternative 
#' approaches to get a token from an environment variable
#' or a file.  
#' 
#'
#' @details 
#' This function will resolve locations of the GDC token in the 
#' following order:
#' \itemize{
#' \item{from the environment variable, \code{GDC_TOKEN}, expected to 
#' contain the token downloaded from the GDC as a string}
#' \item{using \code{readLines} to read a file named in the environment
#' variable, \code{GDC_TOKEN_FILE}}
#' \item{using \code{readLines} to read from a file called \code{.gdc_token} in the user's
#' home directory}
#' }
#' If all of these fail, this function will return an error.
#' 
#' @return character(1) (invisibly, to protect against inadvertently printing) the GDC token.
#' 
#' @references \url{https://docs.gdc.cancer.gov/Data_Portal/Users_Guide/Cart/#gdc-authentication-tokens}
#'
#' @examples 
#' # This will not run before a GDC token
#' # is in place.  
#' token = try(gdc_token(),silent=TRUE)
#' 
#'
#' @export
gdc_token <- function() {
  if(Sys.getenv('GDC_TOKEN')!='') return(Sys.getenv('GDC_TOKEN'))
  token_file = "~/.gdc_token"
  if(Sys.getenv('GDC_TOKEN_FILE')!='') 
    token_file = trimws(Sys.getenv('GDC_TOKEN_FILE'))
  stopifnot(file.exists(token_file))
  invisible(suppressWarnings(readLines(token_file,n=1)))
}

#' Query GDC for data slices
#'
#' This function returns a BAM file representing reads overlapping
#' regions specified either as chromosomal regions or as gencode gene
#' symbols.
#'
#' @param uuid character(1) identifying the BAM file resource
#'
#' @param regions character() vector describing chromosomal regions,
#'     e.g., \code{c("chr1", "chr2:10000", "chr3:10000-20000")} (all
#'     of chromosome 1, chromosome 2 from position 10000 to the end,
#'     chromosome 3 from 10000 to 20000).
#'
#' @param symbols character() vector of gencode gene symbols, e.g.,
#'     \code{c("BRCA1", "PTEN")}
#'
#' @param destination character(1) default \code{tempfile()} file path
#'     for BAM file slice
#'
#' @param overwrite logical(1) default FALSE can destination be
#'     overwritten?
#'
#' @param progress logical(1) default \code{interactive()} should a
#'     progress bar be used?
#'
#' @param token character(1) security token allowing access to
#'     restricted data. Almost all BAM data is restricted, so a token is
#'     usually required. See
#'     \url{https://docs.gdc.cancer.gov/Data/Data_Security/Data_Security/#authentication-tokens}.
#'
#' @details This function uses the Genomic Data Commons "slicing" API
#'     to get portions of a BAM file specified either using "regions"
#'     or using HGNC gene symbols. 
#' 
#' @return character(1) destination to the downloaded BAM file
#'
#' @importFrom httr progress
#' @importFrom jsonlite toJSON
#' 
#' @examples
#' \dontrun{
#'  slicing("df80679e-c4d3-487b-934c-fcc782e5d46e",
#'         regions="chr17:75000000-76000000",
#'         token=gdc_token())
#' 
#' # Get 10 BAM files.
#' bamfiles = files() |> 
#'            filter(data_format=='BAM') |>
#'            results(size=10) |> ids()
#' 
#' # Current alignments at the GDC are to GRCh38
#' library('TxDb.Hsapiens.UCSC.hg38.knownGene')
#' all_genes = genes(TxDb.Hsapiens.UCSC.hg38.knownGene)
#' 
#' first3genes = all_genes[1:3]
#' # remove strand info
#' strand(first3genes) = '*'
#' 
#' # We can get our regions easily now
#' as.character(first3genes)
#' 
#' # Use parallel downloads to speed processing
#' library(BiocParallel)
#' register(MulticoreParam())
#' 
#' fnames = bplapply(bamfiles, slicing, overwrite = TRUE,
#'                 regions=as.character(first3genes))
#' 
#' # 10 BAM files
#' fnames
#' 
#' library(GenomicAlignments)
#' lapply(unlist(fnames), readGAlignments)
#' 
#' }
#' @export
slicing <- function(uuid, regions, symbols, destination=file.path(tempdir(), paste0(uuid, '.bam')),
                    overwrite=FALSE, progress=interactive(), token=gdc_token())
{
    stopifnot(is.character(uuid), length(uuid) == 1L)
    stopifnot(missing(regions) || missing(symbols),
              !(missing(regions) && missing(symbols)))
    stopifnot(is.character(destination), length(destination) == 1L,
              (overwrite && file.exists(destination)) || !file.exists(destination))

    if (!missing(symbols))
        body <- list(gencode=I(symbols))
    else
        ## FIXME: validate regions
        body <- list(regions=regions)

    response <- .gdc_post(
        endpoint=sprintf("slicing/view/%s", uuid),
        add_headers('Content-type'='application/json'),
        write_disk(destination, overwrite),
        if (progress) progress() else NULL,
        body=toJSON(body), token=token)
    if (progress)
        cat("\n")

    destination
}

#' Get the ids associated with a GDC query or response
#'
#' The GDC assigns ids (in the form of uuids) to objects in its database. Those
#' ids can be used for relationships, searching on the website, and as
#' unique ids.  All 
#'
#' @param x A \code{\link{GDCQuery}} or \code{\link{GDCResponse}} object
#'
#' @return a character vector of all the entity ids
#'
#' @examples
#' # use with a GDC query, in this case for "cases"
#' ids(cases() |> filter(~ project.project_id == "TCGA-CHOL"))
#' # also works for responses
#' ids(response(files()))
#' # and results
#' ids(results(cases()))
#'
#' 
#' @export
ids = function(x) {
    UseMethod('ids',x)
}

#' @rdname ids
#' @export
ids.GDCManifest = function(x) {
    return(x[['id']])
}


#' @rdname ids
#' @export
ids.GDCQuery = function(x) {
    fieldname = .id_field(x)
    res = x |> GenomicDataCommons::select(fieldname) |>
        results_all()
    return(.ifNullCharacterZero(res[[fieldname]]))
}


#' @rdname ids
#' @export
ids.GDCResults = function(x) {
    fieldname = .id_field(x)
    res = x[[fieldname]]
    return(.ifNullCharacterZero(res))
}

#' @rdname ids
#' @export
ids.GDCResponse = function(x) {
    fieldname = paste0(sub('s$','',entity_name(x$query)),'_id')
    res = results(x)[[fieldname]]
    return(.ifNullCharacterZero(res))
}

.id_field = function(x) {
    return(paste0(sub('s$','',entity_name(x)),"_id"))
}

#' get the name of the id field
#' 
#' In many places in the GenomicDataCommons package,
#' the entity ids are stored in a column or a vector
#' with a specific name that corresponds to the field name 
#' at the GDC. The format is the entity name (singular) "_id".
#' This generic simply returns that name from a given object.
#' 
#' @param x An object representing the query or results 
#'     of an entity from the GDC ("cases", "files", "annotations", "projects")
#' 
#' @return character(1) such as "case_id", "file_id", etc.
#' 
#' @examples 
#' id_field(cases())
#' 
#' @export
id_field = function(x) {
    UseMethod('id_field',x)
}

#' @describeIn id_field GDCQuery method
#' @export
id_field.GDCQuery = function(x) {
    return(.id_field(x))
}

#' @describeIn id_field GDCResults method
#' @export
id_field.GDCResults = function(x) {
    return(.id_field(x))
}

#' Download GDC files
#'
#' Download one or more files from GDC. Files are downloaded using the
#' UUID and renamed to the file name on the remote system. By default,
#' neither the uuid nor the file name on the remote system can exist.
#'
#' This function is appropriate for one or several files; for large
#' downloads use \code{\link{manifest}} to create a manifest for and
#' the GDC Data Transfer Tool.
#'
#' @param uuids character() of GDC file UUIDs.
#'
#' @param use_cached logical(1) default TRUE indicating that,
#'     if found in the cache, the file will not be downloaded
#'     again. If FALSE, all supplied uuids will be re-downloaded.
#'
#' @param progress logical(1) default TRUE in interactive sessions,
#'     FALSE otherwise indicating whether a progress par should be
#'     produced for each file download.
#'
#' @param access_method character(1), either 'api' or 'client'. See details.
#'
#' @param transfer_args character(1), additional arguments to pass to
#'     the gdc-client command line. See \code{\link{gdc_client}} and
#'     \code{\link{transfer_help}} for details.
#' 
#' @param token (optional) character(1) security token allowing access
#'     to restricted data. See
#'     \url{https://gdc-docs.nci.nih.gov/API/Users_Guide/Authentication_and_Authorization/}.
#'
#' @param ... further arguments passed to files
#'
#' @seealso \code{\link{manifest}} for downloading large data.
#'
#' @return a named vector with file uuids as the names and paths as
#' the value
#'
#' @details When access_method is "api", the GDC "data" endpoint is the
#'     transfer mechanism used. The alternative access_method, "client", will
#'     utilize the \code{gdc-client} transfer tool, which must be
#'     downloaded separately and available. See
#'     \code{\link{gdc_client}} for details on specifying the location
#'     of the gdc-client executable.
#'
#' 
#' @examples
#' # get some example file uuids
#' uuids <- files() |>
#'     filter(~ access == 'open' & file_size < 100000) |>
#'     results(size = 3) |>
#'     ids()
#'
#' # and get the data, placing it into the gdc_cache() directory
#' gdcdata(uuids, use_cached=TRUE)
#'
#' @export
gdcdata <-
    function(uuids, use_cached=TRUE,
             progress=interactive(), token=NULL, access_method='api',
             transfer_args = character(), ...)
{
    stopifnot(is.character(uuids))

    uuids = trimws(uuids)
    manifest = files(...) |>
            GenomicDataCommons::filter( ~ file_id %in% uuids ) |>
            GenomicDataCommons::manifest()
    # files from previous downloads should have the following
    # path and filenames
    fs = file.path(gdc_cache(), manifest[["id"]], manifest[["file_name"]])

    # Restrict new manifest to those that we need to download,
    to_do_manifest = manifest[!file.exists(fs),]

    # These are the uuids of the cache misses
    missing_uuids = to_do_manifest[["id"]]

    # And these are the cache hits
    names(fs) = manifest[["id"]]

    # Using API download to fetch missing uuids
    endpoint <- "data"
    cache_dir <-  gdc_cache()

    destinations <- file.path(cache_dir, missing_uuids)
    if(access_method == 'api') {
        uris <- sprintf("%s/%s", endpoint, missing_uuids)
        value <- mapply(.gdc_download_one, uris, destinations,
                        MoreArgs=list(overwrite=!use_cached, progress=progress,
                                      token=token),
                        SIMPLIFY=TRUE, USE.NAMES=FALSE)
        names(value) <- missing_uuids
    } else {
        ## in the future, may want to transition to
        ## passing the actual manifest, since we
        ## are going to regenerate it, anyway.
        value = NULL
        if(length(missing_uuids)>0) 
            value = transfer(missing_uuids, token = token, args = transfer_args)
    }

    # combine cache hits with cache misses
    #
    # Return vector of file file path, name=uuid
    fs
}

#" (internal) Extract header field element from httr response
#' @importFrom httr headers
.gdc_header_elt <- function(response, field, element) {
    value <- headers(response)[[field]]
    if (is.null(value))
        stop("response header does not contain field '", field, "'")

    value <- strsplit(strsplit(value, "; *")[[1L]], "= *")
    key <- vapply(value, `[[`, character(1), 1L)
    idx <- element == key
    if (sum(idx) != 1L)
        stop("response header field '", field,
             "' does not contain unique element '", element, "'")

    value[[which(idx)]][[2]]
}
    
#" (internal) Rename a file 'from' to 'to'
.gdc_file_rename <- function(from, to, overwrite) {
    if (overwrite && file.exists(to))
        unlink(to)

    reason <- NULL
    status <- withCallingHandlers({
        file.rename(from, to)
    }, warning=function(w) {
        reason <<- conditionMessage(w)
        invokeRestart("muffleWarning")
    })
    unlink(from)
    if (!status)
        stop("failed to rename downloaded file:\n",
             "\n  from: '", from, "'",
             "\n  to: '", to, "'",
             "\n  reason:",
             "\n", .wrapstr(reason))
    else if (!is.null(reason))
        warning(reason)        # forward non-fatal file rename warning

    to
}

#" (internal) GET endpoint / uri
#' @importFrom httr GET add_headers stop_for_status
.gdc_get <-
    function(endpoint, parameters=list(), token=NULL, ..., base=.gdc_base)
{
    stopifnot(is.character(endpoint), length(endpoint) == 1L)
    uri <- sprintf("%s/%s", base, endpoint)
    uri <- sprintf("%s%s", uri, .parameter_string(parameters))
    if(getOption('gdc.verbose',FALSE)) {
      message("GET request uri:\n",uri)
    }
    response <- GET(uri, add_headers(`X-Auth-Token`=token),
                    #config = httr::config(ssl_verifypeer = FALSE),
                    ...)
    stop_for_status(response)
    response
}

#" (internal) POST endpoint / uri
#' @importFrom httr POST add_headers write_disk stop_for_status
.gdc_post <-
    function(endpoint, body, token=NULL, ..., base=.gdc_base)
{
    stopifnot(is.character(endpoint), length(endpoint) == 1L)
    uri <- sprintf("%s/%s", base, endpoint)
    if(getOption('gdc.verbose',FALSE)) {
      message("POST request uri:\n",uri)
      message("POST body: ",jsonlite::toJSON(body))
    }
    if('fields' %in% names(body)) 
        body[['fields']] = paste0(body[['fields']],collapse=',')
    response <- POST(
        uri, add_headers(
            `X-Auth-Token` = token,
            Accept = "application/json",
            `Content-Type` = "application/json"
        ),
        ...,
        #config = httr::config(ssl_verifypeer = FALSE),
        body=body, encode="json")
    stop_for_status(response)
}

#" (internal) Download one file from GDC, renaming to remote filename
#' @importFrom httr GET write_disk add_headers stop_for_status
.gdc_download_one <-
    function(uri, destination, overwrite, progress, token=NULL, base=.gdc_base)
{
    uri = sprintf('%s/%s',base,uri)
    if(getOption('gdc.verbose',FALSE)) {
      message("GET request uri:\n",uri)
    }

    if(!dir.exists(destination)) {
        dir.create(destination)
    }
    destfile = file.path(destination, '.partial_download')
    
    response <- GET(uri, write_disk(destfile, overwrite = TRUE),
                    if (progress) progress() else NULL,
                    add_headers(`X-Auth-Token`=token))
    stop_for_status(response)
    if (progress) cat("\n")
    
    filename <- .gdc_header_elt(response, "content-disposition", "filename")
    to <- file.path(destination, filename)
    .gdc_file_rename(destfile, to, overwrite)
}

.gdc_base <- "https://api.gdc.cancer.gov"
.gdc_endpoint <-
    structure(
        c("status", "projects", "cases", "files", "annotations", "data",
          "manifest", "slicing"), ##, submission
        class="gdc_endpoints")

.gdc_parameters <-
    structure(
        list(format="JSON", pretty=FALSE, fields=NULL, size=10L, from=0L,
             sort=NULL, filters=NULL, facets=NULL),
        class="gdc_parameters")

.gdc_flat_parameters <-
    structure(
        c('fields','facets'),
        class = "gdc_flat_params")

.gdc_entities =
    structure(
        c('projects','cases',"files","annotations", 
          "ssms", "cnvs", "ssm_occurrences", "cnv_occurrences",
          "genes"),
        class = "gdc_entities")

.gdc_manifest_colnames =
    structure(
        c("id", "file_name", "md5sum", "file_size", "state"),
        class = 'gdc_manifest_colnames'
    )


#' Endpoints and Parameters
#'
#' \code{endpoints()} returns available endpoints.
#'
#' @return \code{endpoints()} returns a character vector of possible
#'     endpoints.
#'
#' @rdname constants
#' @examples
#' endpoints()
#' @export
endpoints <- function()
    .gdc_endpoint

#' @export
print.gdc_endpoints <- function(x, ...)
    .cat0("available endpoints:\n", .wrapstr(x), "\n")

#' \code{parameters()} include format (internal use only), pretty
#' (internal use only), fields, size (number of results returned),
#' from (index of rist result), sort, filters, and facets. See
#' \url{https://gdc-docs.nci.nih.gov/API/Users_Guide/Search_and_Retrieval/#query-parameters}
#'
#' @return \code{parameters()} returns a list of possible parameters
#'     and their default values.
#' @keywords internal
#' 
#' @rdname constants
#' @examples
#' parameters()
#' @export
parameters <- function()
    .gdc_parameters

#' @export
print.gdc_parameters <- function(x, ...) {
    cat("available parameters:\n")
    for (nm in names(x))
        .cat0("    ", nm, ": ",
              if (is.null(x[[nm]])) "NULL" else x[[nm]], "\n")
}

#" (internal)
.parameter_string <- function(parameters) {
    if (is.null(parameters))
        return("")
    stopifnot(is.list(parameters),
              all(names(parameters) %in% names(.gdc_parameters)))

    default <- .gdc_parameters
    default[names(parameters)] <- parameters
    default <- Filter(Negate(is.null), default)
    string <- paste(names(default), unname(default), sep="=", collapse="&")
    sprintf("?%s", string)
}

.cat0 <- function(..., sep=NULL)
    cat(..., sep="")

.wrapstr <- function(x)
    paste(strwrap(paste(x, collapse=", "), indent=4, exdent=4), collapse="\n")

.dir_validate_or_create <- function(destination_dir) {
    stopifnot(is.character(destination_dir), length(destination_dir) == 1L,
              nzchar(destination_dir))
    if (!dir.exists(destination_dir)) {
        if (!file.exists(destination_dir))
            dir.create(destination_dir, recursive = TRUE)
        else
            stop("'destination_dir' exists but is not a directory")
    }
}

#" (internal) return character(0) instead of NULL
#"
#" Always return a vector and not NULL.
.ifNullCharacterZero <- function(x) {
    if(is.null(x))
        return(character(0))
    return(x)
}

#' Start a query of GDC metadata
#'
#' The basis for all functionality in this package
#' starts with constructing a query in R. The GDCQuery
#' object contains the filters, facets, and other
#' parameters that define the returned results. A token
#' is required for accessing certain datasets.
#'
#' @aliases GDCQuery
#' 
#' @param entity character vector, including one of the entities in .gdc_entities
#' @param filters a filter list, typically created using \code{\link{make_filter}}, or added
#'     to an existing \code{GDCQuery} object using \code{\link{filter}}.
#' @param facets a character vector of facets for counting common values. 
#'     See \code{\link{available_fields}}. In general, one will not specify this parameter
#'     but will use \code{\link{facet}} instead.
#' @param fields a character vector of fields to return. See \code{\link{available_fields}}.
#'     In general, one will not specify fields directly, but instead use \code{\link{select}}
#' @param expand a character vector of "expands" to include in returned data. See 
#'     \code{\link{available_expand}}
#' 
#' @return An S3 object, the GDCQuery object. This is a list
#' with the following members.
#' \itemize{
#' \item{filters}
#' \item{facets}
#' \item{fields}
#' \item{expand}
#' \item{archive}
#' \item{token}
#' }
#'
#' @examples
#' qcases = query('cases')
#' # equivalent to:
#' qcases = cases()
#' 
#' @export
query = function(entity,
                 filters=NULL,
                 facets=NULL,
                 expand = NULL,
                 fields=default_fields(entity),
                 ...)
{
    stopifnot(entity %in% .gdc_entities)
    ret = structure(
        list(
            fields    = fields,
            filters   = filters,
            facets    = facets,
            expand    = expand),
        class = c(paste0('gdc_',entity),'GDCQuery','list')
    )
    return(ret)
}


#' @describeIn query convenience constructor for a GDCQuery for cases
#'
#' @param ... passed through to \code{\link{query}}
#' 
#' @export
cases = function(...) {return(query('cases',...))}

#' @describeIn query convenience contructor for a GDCQuery for files
#' @export
files = function(...) {return(query('files',...))}

#' @describeIn query convenience contructor for a GDCQuery for projects
#' @export
projects = function(...) {return(query('projects',...))}

#' @describeIn query convenience contructor for a GDCQuery for annotations
#' @export
annotations = function(...) {return(query('annotations',...))}

#' @describeIn query convenience contructor for a GDCQuery for ssms
#' @export
ssms = function(...) {return(query("ssms", ...))}

#' @describeIn query convenience contructor for a GDCQuery for ssm_occurrences
#' @export
ssm_occurrences = function(...) {return(query("ssm_occurrences", ...))}

#' @describeIn query convenience contructor for a GDCQuery for cnvs
#' @export
cnvs = function(...) {return(query("cnvs", ...))}

#' @describeIn query convenience contructor for a GDCQuery for cnv_occurrences
#' @export
cnv_occurrences = function(...) {return(query("cnv_occurrences", ...))}

#' @describeIn query convenience contructor for a GDCQuery for genes
#' @export
genes = function(...) {return(query("genes", ...))}

context('API')

test_that("status returns correctly", {
    res <- status()
    metadata_nms <- c(
        "commit", "data_release", "data_release_version",
        "status", "tag", "version"
    )
    expect_identical(names(res), metadata_nms)
    expect_identical(res$status, "OK")
})

test_that('query', {
    gCases = query('cases')
    expect_equal(class(gCases)[1],'gdc_cases')
    expect_equal(class(gCases)[2],'GDCQuery')
    expect_equal(class(gCases)[3],'list')
    gFiles = query('files')
    expect_equal(class(gFiles)[1],'gdc_files')
    expect_equal(class(gFiles)[2],'GDCQuery')
    expect_equal(class(gFiles)[3],'list')
    gProjects = query('projects')
    expect_equal(class(gProjects)[1],'gdc_projects')
    expect_equal(class(gProjects)[2],'GDCQuery')
    expect_equal(class(gProjects)[3],'list')
    gAnnotations = query('annotations')
    expect_equal(class(gAnnotations)[1],'gdc_annotations')
    expect_equal(class(gAnnotations)[2],'GDCQuery')
    expect_equal(class(gAnnotations)[3],'list')
})

test_that("cases", {
    idfield = "case_id"
    q = cases()
    resp = q |> response()
    expect_gte(q |> count(),1000)
    expect_equal(select(q,idfield)$fields,idfield)
    expect_equal(facet(q,idfield)$facets,idfield)
})

test_that("files", {
    q = files()
    idfield = "file_id"
    resp = q |> response()
    expect_gte(q |> count(),1000)
    expect_equal(select(q,idfield)$fields,idfield)
    expect_equal(facet(q,idfield)$facets,idfield)
})

test_that("annotations", {
    q = annotations()
    idfield = "annotation_id"
    resp = q |> response()
    expect_gte(q |> count(),1000)
    expect_equal(select(q,idfield)$fields,idfield)
    expect_equal(facet(q,idfield)$facets,idfield)
})

test_that("mapping", {
    res = mapping('files')
    expect_equal(class(res),'data.frame')
    expect_equal(ncol(res), 6)
    expect_equal(colnames(res),c('field','description','doc_type','full','type','defaults'))
})

test_that("projects", {
    q = projects()
    idfield = "project_id"
    resp = q |> response()
    expect_gte(q |> count(),35)
    expect_equal(select(q,idfield)$fields,idfield)
    expect_equal(facet(q,idfield)$facets,idfield)
})

library(GenomicDataCommons)
context('data handling')

case_ids <- cases() |> results(size=10) |> ids()

test_that("manifest files", {
    m <- manifest(files(), size = 10)
    expect_identical(nrow(m), 10L)
    expect_true(ncol(m) > 5)
})

test_that("write_manifest", {
    m = files() |> manifest(size=10)
    tf = tempfile()
    write_manifest(m, tf)
    expect_true(file.exists(tf))
    unlink(tf)
})

test_that("gdcdata", {
    d = tempfile()
    if (!dir.exists(d))
        dir.create(d)
    gdc_set_cache(d)
    
    few_file_ids = files() |>
        filter( ~ cases.project.project_id == 'TCGA-SARC' &
            data_type == 'Copy Number Segment' &
            analysis.workflow_type == 'DNAcopy') |> results(size=2) |> ids()

    res = gdcdata(few_file_ids)
    expect_length(res, 2)
    expect_named(res)
    unlink(d, recursive = TRUE)
})

## ----init, results='hide', echo=FALSE, warning=FALSE, message=FALSE-----------
library(knitr)
opts_chunk$set(warning=FALSE, message=FALSE)
BiocStyle::markdown()

## ----install_bioc, eval=FALSE-------------------------------------------------
# if (!require("BiocManager"))
#     install.packages("BiocManager")
# BiocManager::install('GenomicDataCommons')

## ----libraries, message=FALSE-------------------------------------------------
library(GenomicDataCommons)

## ----statusQS-----------------------------------------------------------------
GenomicDataCommons::status()

## ----statusCheck--------------------------------------------------------------
stopifnot(GenomicDataCommons::status()$status=="OK")

## ----manifest-----------------------------------------------------------------
ge_manifest <- files() |>
    filter( cases.project.project_id == 'TCGA-OV') |> 
    filter( type == 'gene_expression' ) |>
    filter( analysis.workflow_type == 'STAR - Counts')  |>
    manifest()
head(ge_manifest)

## ----downloadQS, eval=FALSE---------------------------------------------------
# fnames <- lapply(ge_manifest$id[1:20], gdcdata)

## ----gdc_clinical-------------------------------------------------------------
case_ids = cases() |> results(size=10) |> ids()
clindat = gdc_clinical(case_ids)
names(clindat)

## ----clinData-----------------------------------------------------------------
head(clindat[["main"]])
head(clindat[["diagnoses"]])

## ----metadataQS---------------------------------------------------------------
expands = c("diagnoses","annotations",
             "demographic","exposures")
clinResults = cases() |>
    GenomicDataCommons::select(NULL) |>
    GenomicDataCommons::expand(expands) |>
    results(size=50)
str(clinResults[[1]],list.len=6)
# or listviewer::jsonedit(clinResults)

## ----projectquery-------------------------------------------------------------
pquery = projects()

## ----pquery-------------------------------------------------------------------
str(pquery)

## ----pquerycount--------------------------------------------------------------
pcount = count(pquery)
# or
pcount = pquery |> count()
pcount

## ----pqueryresults------------------------------------------------------------
presults = pquery |> results()

## ----presultsstr--------------------------------------------------------------
str(presults)

## ----presultsall--------------------------------------------------------------
length(ids(presults))
presults = pquery |> results_all()
length(ids(presults))
# includes all records
length(ids(presults)) == count(pquery)

## ----defaultfields------------------------------------------------------------
default_fields('files')
# The number of fields available for files endpoint
length(available_fields('files'))
# The first few fields available for files endpoint
head(available_fields('files'))

## ----selectexample------------------------------------------------------------
# Default fields here
qcases = cases()
qcases$fields
# set up query to use ALL available fields
# Note that checking of fields is done by select()
qcases = cases() |> GenomicDataCommons::select(available_fields('cases'))
head(qcases$fields)

## ----aggexample---------------------------------------------------------------
# total number of files of a specific type
res = files() |> facet(c('type','data_type')) |> aggregations()
res$type

## ----allfilesunfiltered-------------------------------------------------------
qfiles = files()
qfiles |> count() # all files

## ----onlyGeneExpression-------------------------------------------------------
qfiles = files() |> filter( type == 'gene_expression')
# here is what the filter looks like after translation
str(get_filter(qfiles))

## ----filtAvailFields----------------------------------------------------------
grep('pro',available_fields('files'),value=TRUE) |> 
    head()

## ----filtProgramID------------------------------------------------------------
files() |> 
    facet('cases.project.project_id') |> 
    aggregations() |> 
    head()

## ----filtfinal----------------------------------------------------------------
qfiles = files() |>
    filter( cases.project.project_id == 'TCGA-OV' & type == 'gene_expression')
str(get_filter(qfiles))
qfiles |> count()

## ----filtChain----------------------------------------------------------------
qfiles2 = files() |>
    filter( cases.project.project_id == 'TCGA-OV') |> 
    filter( type == 'gene_expression') 
qfiles2 |> count()
(qfiles |> count()) == (qfiles2 |> count()) #TRUE

## ----filtAndManifest----------------------------------------------------------
manifest_df = qfiles |> manifest()
head(manifest_df)

## ----filterForSTARCounts------------------------------------------------------
qfiles = files() |> filter( ~ cases.project.project_id == 'TCGA-OV' &
                            type == 'gene_expression' &
                            access == "open" &
                            analysis.workflow_type == 'STAR - Counts')
manifest_df = qfiles |> manifest()
nrow(manifest_df)

## ----authenNoRun, eval=FALSE--------------------------------------------------
# token = gdc_token()
# transfer(...,token=token)
# # or
# transfer(...,token=get_token())

## ----singlefileDL-------------------------------------------------------------
fnames = gdcdata(manifest_df$id[1:2],progress=FALSE)


## ----bulkDL, eval=FALSE-------------------------------------------------------
# # Requires gcd_client command-line utility to be isntalled
# # separately.
# fnames = gdcdata(manifest_df$id[3:10], access_method = 'client')

## ----casesPerProject----------------------------------------------------------
res = cases() |> facet("project.project_id") |> aggregations()
head(res)
library(ggplot2)
ggplot(res$project.project_id,aes(x = key, y = doc_count)) +
    geom_bar(stat='identity') +
    theme(axis.text.x = element_text(angle = 45, hjust = 1))

## ----casesInTCGA--------------------------------------------------------------
cases() |> filter(~ project.program.name=='TARGET') |> count()

## ----casesInTARGET------------------------------------------------------------
cases() |> filter(~ project.program.name=='TCGA') |> count()

## ----casesTCGABRCASampleTypes-------------------------------------------------
# The need to do the "&" here is a requirement of the
# current version of the GDC API. I have filed a feature
# request to remove this requirement.
resp = cases() |> filter(~ project.project_id=='TCGA-BRCA' &
                              project.project_id=='TCGA-BRCA' ) |>
    facet('samples.sample_type') |> aggregations()
resp$samples.sample_type

## ----casesTCGABRCASolidNormal-------------------------------------------------
# The need to do the "&" here is a requirement of the
# current version of the GDC API. I have filed a feature
# request to remove this requirement.
resp = cases() |> filter(~ project.project_id=='TCGA-BRCA' &
                              samples.sample_type=='Solid Tissue Normal') |>
    GenomicDataCommons::select(c(default_fields(cases()),'samples.sample_type')) |>
    response_all()
count(resp)
res = resp |> results()
str(res[1],list.len=6)
head(ids(resp))

## ----casesFemaleTCGA----------------------------------------------------------
cases() |>
  GenomicDataCommons::filter(~ project.program.name == 'TCGA' &
    "cases.demographic.gender" %in% "female") |>
      GenomicDataCommons::results(size = 4) |>
        ids()

## ----notFemaleTCGACOAD--------------------------------------------------------
cases() |>
  GenomicDataCommons::filter(~ project.project_id == 'TCGA-COAD' &
    "cases.demographic.gender" %exclude% "female") |>
      GenomicDataCommons::results(size = 4) |>
        ids()

## ----missingGenderTCGA--------------------------------------------------------
cases() |>
  GenomicDataCommons::filter(~ project.program.name == 'TCGA' &
    missing("cases.demographic.gender")) |>
      GenomicDataCommons::results(size = 4) |>
        ids()

## ----notMissingGenderTCGA-----------------------------------------------------
cases() |>
  GenomicDataCommons::filter(~ project.program.name == 'TCGA' &
    !missing("cases.demographic.gender")) |>
      GenomicDataCommons::results(size = 4) |>
        ids()

## ----filesVCFCount------------------------------------------------------------
res = files() |> facet('type') |> aggregations()
res$type
ggplot(res$type,aes(x = key,y = doc_count)) + geom_bar(stat='identity') +
    theme(axis.text.x = element_text(angle = 45, hjust = 1))

## ----filesRNAseqGeneGBM-------------------------------------------------------
q = files() |>
    GenomicDataCommons::select(available_fields('files')) |>
    filter(~ cases.project.project_id=='TCGA-GBM' &
               data_type=='Gene Expression Quantification')
q |> facet('analysis.workflow_type') |> aggregations()
# so need to add another filter
file_ids = q |> filter(~ cases.project.project_id=='TCGA-GBM' &
                            data_type=='Gene Expression Quantification' &
                            analysis.workflow_type == 'STAR - Counts') |>
    GenomicDataCommons::select('file_id') |>
    response_all() |>
    ids()

## ----filesRNAseqGeneGBMforBAM-------------------------------------------------
q = files() |>
    GenomicDataCommons::select(available_fields('files')) |>
    filter(~ cases.project.project_id == 'TCGA-GBM' &
               data_type == 'Aligned Reads' &
               experimental_strategy == 'RNA-Seq' &
               data_format == 'BAM')
file_ids = q |> response_all() |> ids()

## ----slicing10, eval=FALSE----------------------------------------------------
# bamfile = slicing(file_ids[1],regions="chr12:6534405-6538375",token=gdc_token())
# library(GenomicAlignments)
# aligns = readGAlignments(bamfile)

## ----sessionInfo--------------------------------------------------------------
sessionInfo()

.response_mapping_as_list <- function(json) {
    json <- lapply(json, unlist)
    structure(json, class=c("mapping_list", "gdc_list", "list"))
}


#" (internal) utility for returning _mapping json
#' @importFrom httr content
.get_mapping_json <- function(endpoint) {
    valid <- .gdc_entities
    stopifnot(is.character(endpoint), length(endpoint) == 1L,
              endpoint %in% valid)
    response <- .gdc_get(
        sprintf("%s/%s", endpoint, "_mapping")
    )
    content(response, type="application/json")
}


#' Query GDC for available endpoint fields
#'
#' @param endpoint character(1) corresponding to endpoints for which
#'     users may specify additional or alternative fields. Endpoints
#'     include \dQuote{projects}, \dQuote{cases}, \dQuote{files}, and
#'     \dQuote{annotations}.
#'
#' @return A data frame describing the field (field name), full (full
#'     data model name), type (data type), and four additional columns
#'     describing the "set" to which the fields belong--\dQuote{default},
#'     \dQuote{expand}, \dQuote{multi}, and \dQuote{nested}.
#'
#' @examples
#' map <- mapping("projects")
#' head(map)
#' # get only the "default" fields
#' subset(map,defaults)
#' # And get just the text names of the "default" fields
#' subset(map,defaults)$field
#' 
#' @importFrom httr content
#' @export
mapping <- function(endpoint) {
    json = .get_mapping_json(endpoint)
    maplist = list()
    fields = data.frame(field=unlist(json[['fields']]))
    mapdat = json[['_mapping']]
    for(cname in names(mapdat[[1]])) {
        maplist[[cname]] = as.character(sapply(mapdat,'[[',cname))
    }
    df = do.call(cbind,maplist)
    tmpdf = as.data.frame(matrix(FALSE, ncol = 1, nrow = nrow(df)),stringsAsFactors = FALSE)
    fieldtypes = c('defaults')
    colnames(tmpdf) = fieldtypes
    df = cbind(data.frame(df,stringsAsFactors=FALSE),tmpdf)
    df = as.data.frame(merge(fields,df,by.x='field',by.y='field',all.x=TRUE),stringsAsFactors = FALSE)
    df$field = as.character(df$field)
    for(i in fieldtypes) {
        df[df$field  %in% json[[i]],i] = TRUE
    }
    return(df)
}

#' Read DNAcopy results into GRanges object
#'
#' @param fname The path to a DNAcopy-like file.
#' @param ... passed to \code{\link[readr]{read_tsv}}
#' @return a \code{\link[GenomicRanges]{GRanges}} object
#' 
#' @importFrom readr read_tsv
#' @import GenomicRanges
#' @importFrom IRanges IRanges
#'
#' @examples
#' fname = system.file(package='GenomicDataCommons',
#'                     'extdata/dnacopy.tsv.gz')
#' dnac = readDNAcopy(fname)
#' class(dnac)
#' length(dnac)
#'
#' @export
readDNAcopy <- function(fname,...) {
    stopifnot(file.exists(fname))
    res = read_tsv(fname,...)
    stopifnot(ncol(res)==6)
    return(GRanges(seqnames=res[[2]],
                   ranges=IRanges(start=res[[3]],end=res[[4]]),
                   sampleName = res[[1]],
                   Num_Probes = res[[5]],
                   value      = res[[6]]))
}

#' Return valid values for "expand"
#'
#' The GDC allows a shorthand for specifying groups
#' of fields to be returned by the metadata queries.
#' These can be specified in a \code{\link{select}}
#' method call to easily supply groups of fields.
#'
#' @param entity Either a \code{\link{GDCQuery}} object
#' or a character(1) specifying a GDC entity ('cases', 'files',
#' 'annotations', 'projects')
#'
#' @return A character vector
#' 
#' @seealso See \url{https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#expand}
#' for details
#'
#' @examples
#' head(available_expand('files'))
#'
#' @export
available_expand <- function(entity) {
    UseMethod("available_expand",entity)
}   

#' @rdname available_expand
#'
#' @export
available_expand.character <- function(entity) {
    json = .get_mapping_json(entity)
    return(unlist(json[['expand']]))
}

#' @rdname available_expand
#'
#' @export
available_expand.GDCQuery <- function(entity) {
    return(available_expand(entity_name(entity)))
}

#" (internal) check expand values
.gdcCheckExpands <- function(entity,expand) {
    if(is.null(expand)) return(TRUE)
    stopifnot(entity %in% .gdc_entities)
    ae = available_expand(entity)
    mismatches = expand[!(expand %in% ae)]
    if(length(mismatches)>0)
        stop(sprintf('expand specified included expands not available in %s including (%s)',entity,mismatches))
    return(TRUE)
}

#' Set the \code{expand} parameter
#'
#' S3 generic to set GDCQuery expand parameter
#'
#' @param x the objects on which to set fields
#' @param expand a character vector specifying the fields
#' 
#'
#' @return A \code{\link{GDCQuery}} object, with the \code{expand}
#' member altered.
#' 
#' @examples
#' gProj = projects()
#' gProj$fields
#' head(available_fields(gProj))
#' default_fields(gProj)
#'
#' gProj |>
#'   select(default_fields(gProj)[1:2]) |>
#'   response() |>
#'   str(max_level=2)
#' 
#' @export
expand <- function(x,expand) {
    UseMethod('expand',x)
}

#' @describeIn expand set expand fields on a GDCQuery object
#' @export
expand.GDCQuery <- function(x,expand) {
    .gdcCheckExpands(entity_name(x),expand)
    x$expand = expand
    return(x)
}

#' Prepare GDC manifest file for bulk download
#'
#' The \code{manifest} function/method creates a manifest of files to be downloaded
#' using the GDC Data Transfer Tool. There are methods for
#' creating manifest data frames from \code{\link{GDCQuery}} objects
#' that contain file information ("cases" and "files" queries).
#' 
#' @param x An \code{\link{GDCQuery}} object of subclass "gdc_files" or "gdc_cases".
#'
#' @param size The total number of records to return.  Default 
#' will return the usually desirable full set of records.
#'
#' @param from Record number from which to start when returning the manifest.
#'
#' @param ... passed to \code{\link[httr]{PUT}}.
#'
#' @return A \code{\link[tibble]{tibble}}, also of type "gdc_manifest", with five columns:
#' \itemize{
#' \item{id}
#' \item{filename}
#' \item{md5}
#' \item{size}
#' \item{state}
#'}
#' 
#' @examples
#' gFiles = files()
#' shortManifest = gFiles |> manifest(size=10)
#' head(shortManifest,n=3)
#'
#' 
#' @export
manifest <- function(x,from=0,size=count(x),...) {
    UseMethod('manifest',x)
}

#' @describeIn manifest
#'
#' @export
manifest.gdc_files <- function(x,from=0,size=count(x),...) {
    .manifestCall(x=x,from=from,size=size,...)
}

#' @describeIn manifest
#'
#' @export
manifest.GDCfilesResponse <- function(x,from=0,size=count(x),...) {
    .manifestCall(x=x$query,from=from,size=size,...)
}

#' @describeIn manifest
#'
#' @export
manifest.GDCcasesResponse <- function(x,from=0,size=count(x),...) {
    manifest(x=x$query,from=from,size=size,...)
}



#' @importFrom readr read_tsv 
.manifestCall <- function(x,from=0,size=count(x),...) {
    body = Filter(function(z) !is.null(z),x)
    body[['facets']]=NULL
    body[['fields']]=paste0(default_fields(x),collapse=",")
    body[['from']]=from
    body[['size']]=size
    # remove return_type for now
    # body[['return_type']]='manifest'
    tmp <- httr::content(
        .gdc_post(entity_name(x), body=body, token=NULL, ...),
        as = "text", encoding = "UTF-8"
    )
    tmp <- jsonlite::fromJSON(tmp)[["data"]][["hits"]]
    if ("acl" %in% names(tmp))
        tmp <- tidyr::unnest_wider(data = tmp, col = "acl", names_sep = "_")
    if(ncol(tmp)<5) {
        tmp=data.frame()
    }
    class(tmp) <- c('GDCManifest',class(tmp))
    return(tmp)
}

#' write a manifest data.frame to disk
#' 
#' The \code{\link{manifest}} method creates a data.frame
#' that represents the data for a manifest file needed
#' by the GDC Data Transfer Tool. While the file format
#' is nothing special, this is a simple helper function
#' to write a manifest data.frame to disk. It returns
#' the path to which the file is written, so it can
#' be used "in-line" in a call to \code{\link{transfer}}.
#' 
#' @param manifest A data.frame with five columns, typically
#'     created by a call to \code{\link{manifest}}
#' 
#' @param destfile The filename for saving the manifest.
#'          
#' @return character(1) the destination file name.
#'
#' @importFrom utils write.table
#' 
#' @examples
#' mf = files() |> manifest(size=10)
#' write_manifest(mf)
#' 
#' @export
write_manifest <- function(manifest,destfile=tempfile()) {
    stopifnot(
        all(.gdc_manifest_colnames %in% colnames(manifest)),
        ncol(manifest) > 5
    )
    write.table(manifest,file=destfile,sep="\t",
                col.names=TRUE,row.names=FALSE,quote=FALSE)
    destfile
}

library(httr)
library(xml2)

pkghome <- "~/a/GenomicDataCommons"

url <-
    "https://gdc-docs.nci.nih.gov/API/Users_Guide/Appendix_A_Available_Fields/"
xml = content(GET(url))

.get_field <- function(xml, xpath) {
    fields <- as.character(xml_find_all(xml, xpath))
    Filter(nzchar, trimws(fields))
}

.project_fields <- .get_field(xml, "//table[1]//tr/td[1]/text()")
.file_fields <- .get_field(xml, "//table[2]//tr/td[1]/text()")
.case_fields <- .get_field(xml, "//table[3]//tr/td[1]/text()")
.annotation_fields <- .get_field(xml, "//table[4]//tr/td[1]/text()")

save(.project_fields, .file_fields, .case_fields, .annotation_fields,
     file=file.path(pkghome, "R", "sysdata.rda"))

#.unary_op <- function(left) {
#  force(left)
#  function(e1) {
#    force(e1)
#    list(op=e1,content=c(field=left,value=c(right)))
#  }
#}

#' @importFrom jsonlite unbox
.binary_op <- function(sep) {
  force(sep)
  function(e1, e2) {
      force(e1)
      force(e2)
      list(op=unbox(sep),content=list(field=e1,value=e2))
  }
}

.missing_op <- function(sep) {
  force(sep)
  function(e1) {
      force(e1)
      list(op=unbox(sep), content = list(field = e1, value = "MISSING"))
  }
}

.negate_op <- function(sep) {
    force(sep)
    function(op) {
      force(op)
      list(op = unbox(sep),
        content = list(
          field = op$content$field,
          value = op$content$value)
      )
    }
}

#' @importFrom jsonlite unbox
.combine_op <- function(sep) {
  force(sep)
  function(e1, e2) {
      force(e1)
      force(e2)
    return(list(op=unbox(sep),content=list(e1,e2)))
  }
}

#.f_env = new.env(parent=emptyenv())
.f_env = list()
.f_env$`==` = .binary_op('=')
.f_env$`!=` = .binary_op('!=')
.f_env$`<` = .binary_op('<')
.f_env$'>' = .binary_op('>')
.f_env$'&' = .combine_op('and')
.f_env$'|' = .combine_op('or')
.f_env$`<=` = .binary_op('<=')
.f_env$'>=' = .binary_op('>=')
.f_env$'%in%' = .binary_op('in')
.f_env$'%exclude%' = .binary_op('exclude')
.f_env$`missing` = .missing_op("is")
.f_env$'!' = .negate_op("NOT")


#' Create NCI GDC filters for limiting GDC query results
#'
#' Searching the NCI GDC allows for complex filtering based
#' on logical operations and simple comparisons.  This function
#' facilitates writing such filter expressions in R-like syntax
#' with R code evaluation.
#'
#' If used with available_fields, "bare" fields that are
#' named in the available_fields character vector can be used
#' in the filter expression without quotes.
#'
#' @param expr a lazy-wrapped expression or a formula RHS equivalent
#'
#' @param available_fields a character vector of the
#' additional names that will be injected into the
#' filter evaluation environment
#'
#' @return a \code{list} that represents an R version
#' of the JSON that will ultimately be used in an
#' NCI GDC search or other query.
#' 
#' @importFrom rlang eval_tidy f_rhs f_env
#' 
#' @export
make_filter = function(expr,available_fields) {
    available_fields=as.list(available_fields)
    names(available_fields)=available_fields
    filt_env = c(as.list(.f_env),available_fields)
    if(is_formula(expr)) {
        return(rlang::eval_tidy(rlang::f_rhs(expr), data=filt_env, env = rlang::f_env(expr)))
    } else {
        return(rlang::eval_tidy(expr,data=filt_env))
    }
}



#' Manipulating GDCQuery filters
#'
#' @name filtering
#' 
#' @return A \code{\link{GDCQuery}} object with the filter
#' field replaced by specified filter expression
#' 
#' @examples
#' # make a GDCQuery object to start
#' #
#' # Projects
#' #
#' pQuery = projects()
#'
#' # check for the default fields
#' # so that we can use one of them to build a filter
#' default_fields(pQuery)
#' pQuery = filter(pQuery,~ project_id == 'TCGA-LUAC')
#' get_filter(pQuery)
#'
#' #
#' # Files
#' #
#' fQuery = files()
#' default_fields(fQuery)
#'
#' fQuery = filter(fQuery,~ data_format == 'VCF')
#' # OR
#' # with recent GenomicDataCommons versions:
#' #   no "~" needed
#' fQuery = filter(fQuery, data_format == 'VCF')
#' 
#' get_filter(fQuery)
#'
#' fQuery = filter(fQuery,~ data_format == 'VCF'
#'                 & experimental_strategy == 'WXS'
#'                 & type == 'simple_somatic_mutation')
#'
#' files() |> filter(~ data_format == 'VCF'
#'                    & experimental_strategy=='WXS'
#'                    & type == 'simple_somatic_mutation') |> count()
#'                    
#'                    
#' files() |> filter( data_format == 'VCF'
#'                    & experimental_strategy=='WXS'
#'                    & type == 'simple_somatic_mutation') |> count()
#'
#' # Filters may be chained for the 
#' # equivalent query
#' # 
#' # When chained, filters are combined with logical AND
#' 
#' files() |>
#'   filter(~ data_format == 'VCF') |>
#'   filter(~ experimental_strategy == 'WXS') |>
#'   filter(~ type == 'simple_somatic_mutation') |>
#'   count()
#' 
#' # OR
#' 
#' files() |>
#'   filter( data_format == 'VCF') |>
#'   filter( experimental_strategy == 'WXS') |>
#'   filter( type == 'simple_somatic_mutation') |>
#'   count()
#' 
#' # Use str() to get a cleaner picture
#' str(get_filter(fQuery))
NULL

#' The \code{filter} is simply a safe accessor for
#' the filter element in \code{\link{GDCQuery}} objects.
#'
#' @param x the object on which to set the filter list
#' member
#' @param expr a filter expression in the form of
#' the right hand side of a formula, where bare names
#' (without quotes) are allowed if they are available
#' fields associated with the GDCQuery object, \code{x}
#' 
#' @rdname filtering
#' 
#' @export
filter = function(x,expr) {
    UseMethod('filter',x)
}

#' @rdname filtering
#'
#' @importFrom rlang enquo is_formula
#'
#' @export
filter.GDCQuery = function(x,expr) {
    filt = try({
        if(rlang::is_formula(expr))
            make_filter(expr,available_fields(x))
    }, silent=TRUE)
    if(inherits(filt, "try-error")) 
        filt = make_filter(enquo(expr), available_fields(x))
    if(!is.null(x$filters))
        x$filters=list(op="and", content=list(x$filters,filt))
    else
        x$filters = filt
    return(x)
}

#' The \code{get_filter} is simply a safe accessor for
#' the filter element in \code{\link{GDCQuery}} objects.
#'
#' @rdname filtering
#'
#' 
#' @export
get_filter = function(x) {
    UseMethod('get_filter',x)
}

#' @rdname filtering
#' 
#' @export
get_filter.GDCQuery = function(x) {
    return(x$filters)
}

#' Set facets for a \code{\link{GDCQuery}}
#'
#' @param x a \code{\link{GDCQuery}} object
#' @param facets a character vector of fields that
#' will be used for forming aggregations (facets).
#' Default is to set facets for all default fields.
#' See \code{\link{default_fields}} for details
#'
#' @return returns a \code{\link{GDCQuery}} object,
#' with facets field updated.
#' 
#' @rdname faceting
#'
#' @examples
#' # create a new GDCQuery against the projects endpoint
#' gProj = projects()
#'
#' # default facets are NULL
#' get_facets(gProj)
#'
#' # set facets and save result
#' gProjFacet = facet(gProj)
#'
#' # check facets
#' get_facets(gProjFacet)
#' 
#' # and get a response, noting that
#' # the aggregations list member contains
#' # tibbles for each facet
#' str(response(gProjFacet,size=2),max.level=2)
#' 
#' @export
facet = function(x,facets) {
    UseMethod('facet',x)
}


#' @export
facet.GDCQuery = function(x,facets=default_fields(x)) {
    x$facets = facets
    return(x)
}

#' Get facets for a \code{\link{GDCQuery}}
#'
#' @rdname faceting
#' 
#' @export
get_facets = function(x) {
    UseMethod('get_facets',x)
}

#' @rdname faceting
#'
#' @export
get_facets.GDCQuery = function(x) {
    return(x$facets)
}

---
title: "Questions and answers from over the years"
author: "Sean Davis"
date: "`r format(Sys.Date(), '%A, %B %d, %Y')`"
always_allow_html: yes
output:
  BiocStyle::html_document:
    df_print: paged
    toc_float: true
    keep_md: true
abstract: >

vignette: >
  %\VignetteIndexEntry{Questions and answers from over the years}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# How could I generate a manifest file with filtering of Race and Ethnicity?

From https://support.bioconductor.org/p/9138939/.

```{r}
library(GenomicDataCommons,quietly = TRUE)
```

I made a small change to the filtering expression approach based on 
changes to lazy evaluation best practices. There is now no need to 
include the `~` in the filter expression. So:

```{r}
q = files() |>
  GenomicDataCommons::filter(
    cases.project.project_id == 'TCGA-COAD' &
      data_type == 'Aligned Reads' &
      experimental_strategy == 'RNA-Seq' &
      data_format == 'BAM')
```
And get a count of the results:

```{r}
count(q)
```

And the manifest.

```{r}
manifest(q)
```

Your question about race and ethnicity is a good one. 

```{r}
all_fields = available_fields(files())
```

And we can grep for `race` or `ethnic` to get potential matching fields
to look at.

```{r}
grep('race|ethnic',all_fields,value=TRUE)
```

Now, we can check available values for each field to determine how to complete
our filter expressions.

```{r}
available_values('files',"cases.demographic.ethnicity")
available_values('files',"cases.demographic.race")
```

We can complete our filter expression now to limit to `white` race only.

```{r}
q_white_only = q |>
  GenomicDataCommons::filter(cases.demographic.race=='white')
count(q_white_only)
manifest(q_white_only)
```

# How can I get the number of cases with RNA-Seq data added by date to TCGA project with `GenomicDataCommons`?

- From https://support.bioconductor.org/p/9135791/

I would like to get the number of cases added (created, any logical datetime would suffice here) to the TCGA project by experiment type. I attempted to get this data via GenomicDataCommons package, but it is giving me I believe the number of files for a given experiment type rather than number cases. How can I get the number of cases for which there is RNA-Seq data?

```{r}
library(tibble)
library(dplyr)
library(GenomicDataCommons)

cases() |> 
  GenomicDataCommons::filter(
    ~ project.program.name=='TCGA' & files.experimental_strategy=='RNA-Seq'
  ) |> 
  facet(c("files.created_datetime")) |> 
  aggregations() |> 
  unname() |>
  unlist(recursive = FALSE) |> 
  as_tibble() |>
  dplyr::arrange(dplyr::desc(key))
```

---
title: "The GenomicDataCommons Package"
author: "Sean Davis & Martin Morgan"
date: "`r format(Sys.Date(), '%A, %B %d, %Y')`"
always_allow_html: yes
output:
  BiocStyle::html_document:
    df_print: paged
    toc_float: true
abstract: >
  The National Cancer Institute (NCI) has established
  the [Genomic Data Commons](https://gdc.nci.nih.gov/) (GDC). The GDC
  provides the cancer research community with an open and unified
  repository for sharing and accessing data across numerous cancer
  studies and projects via a high-performance data transfer and query
  infrastructure.  The *GenomicDataCommons* Bioconductor package
  provides basic infrastructure for querying, accessing, and mining
  genomic datasets available from the GDC. We expect that the
  Bioconductor developer and the larger bioinformatics communities will
  build on the *GenomicDataCommons* package to add higher-level
  functionality and expose cancer genomics data to the plethora of
  state-of-the-art bioinformatics methods available in Bioconductor.

vignette: >
  %\VignetteIndexEntry{Introduction to Accessing the NCI Genomic Data Commons}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

```{r init, results='hide', echo=FALSE, warning=FALSE, message=FALSE}
library(knitr)
opts_chunk$set(warning=FALSE, message=FALSE)
BiocStyle::markdown()
```


# What is the GDC?

From the [Genomic Data Commons (GDC) website](https://gdc.cancer.gov/about-gdc):

> The National Cancer Institute's (NCI's) Genomic Data Commons (GDC) is
a data sharing platform that promotes precision medicine in
oncology. It is not just a database or a tool; it is an expandable
knowledge network supporting the import and standardization of genomic
and clinical data from cancer research programs.
> The GDC contains NCI-generated data from some of the largest and most
comprehensive cancer genomic datasets, including The Cancer Genome
Atlas (TCGA) and Therapeutically Applicable Research to Generate
Effective Therapies (TARGET). For the first time, these datasets have
been harmonized using a common set of bioinformatics pipelines, so
that the data can be directly compared.
> As a growing knowledge system for cancer, the GDC also enables
researchers to submit data, and harmonizes these data for import into
the GDC. As more researchers add clinical and genomic data to the GDC,
it will become an even more powerful tool for making discoveries about
the molecular basis of cancer that may lead to better care for
patients.

The
[data model for the GDC is complex](https://gdc.cancer.gov/developers/gdc-data-model/gdc-data-model-components),
but it worth a quick overview and a graphical representation is included here. 

![The data model is encoded as a
so-called property graph. Nodes represent entities such as Projects,
Cases, Diagnoses, Files (various kinds), and Annotations. The
relationships between these entities are maintained as edges.  Both
nodes and edges may have Properties that supply instance details. ](all_nodes_040318.png)

 The
GDC API exposes these nodes and edges in a somewhat simplified set
of
[RESTful](https://en.wikipedia.org/wiki/Representational_state_transfer) endpoints.

# Quickstart

This quickstart section is just meant to show basic
functionality. More details of functionality are included further on
in this vignette and in function-specific help.

This software is available at Bioconductor.org and can be downloaded via
`BiocManager::install`.

To report bugs or problems, either
[submit a new issue](https://github.com/Bioconductor/GenomicDataCommons/issues)
or submit a `bug.report(package='GenomicDataCommons')` from within R (which
will redirect you to the new issue on GitHub).

## Installation

Installation can be achieved via Bioconductor's `BiocManager` package.

```{r install_bioc, eval=FALSE}
if (!require("BiocManager"))
    install.packages("BiocManager")
BiocManager::install('GenomicDataCommons')
```

```{r libraries, message=FALSE}
library(GenomicDataCommons)
```

## Check connectivity and status

The `r Biocpkg("GenomicDataCommons")` package relies on having network
connectivity. In addition, the NCI GDC API must also be operational
and not under maintenance. Checking `status` can be used to check this
connectivity and functionality.

```{r statusQS}
GenomicDataCommons::status()
```

And to check the status in code:

```{r statusCheck}
stopifnot(GenomicDataCommons::status()$status=="OK")
```


## Find data

The following code builds a `manifest` that can be used to guide the
download of raw data. Here, filtering finds gene expression files
quantified as raw counts using `STAR` from ovarian cancer patients.

```{r manifest}
ge_manifest <- files() |>
    filter( cases.project.project_id == 'TCGA-OV') |> 
    filter( type == 'gene_expression' ) |>
    filter( analysis.workflow_type == 'STAR - Counts')  |>
    manifest()
head(ge_manifest)
```

## Download data

After  the `r nrow(ge_manifest)` gene expression files
specified in the query above. Using multiple processes to do the download very
significantly speeds up the transfer in many cases.  On a standard 1Gb
connection, the following completes in about 30 seconds. The first time the 
data are downloaded, R will ask to create a cache directory (see `?gdc_cache`
for details of setting and interacting with the cache). Resulting
downloaded files will be stored in the cache directory. Future access to 
the same files will be directly from the cache, alleviating multiple downloads.

```{r downloadQS, eval=FALSE}
fnames <- lapply(ge_manifest$id[1:20], gdcdata)
```

If the download had included controlled-access data, the download above would
have needed to include a `token`.  Details are available in
[the authentication section below](#authentication).

## Metadata queries

### Clinical data

Accessing clinical data is a very common task. Given a set of `case_ids`,
the `gdc_clinical()` function will return a list of four `tibble`s. 

- demographic
- diagnoses
- exposures
- main

```{r gdc_clinical}
case_ids = cases() |> results(size=10) |> ids()
clindat = gdc_clinical(case_ids)
names(clindat)
```

```{r clinData}
head(clindat[["main"]])
head(clindat[["diagnoses"]])
```

### General metadata queries

The `r Biocpkg("GenomicDataCommons")` package can access the significant 
clinical, demographic, biospecimen, and annotation information 
contained in the NCI GDC. The `gdc_clinical()` function will often
be all that is needed, but the API and `r Biocpkg("GenomicDataCommons")` package
make much flexibility if fine-tuning is required. 

```{r metadataQS}
expands = c("diagnoses","annotations",
             "demographic","exposures")
clinResults = cases() |>
    GenomicDataCommons::select(NULL) |>
    GenomicDataCommons::expand(expands) |>
    results(size=50)
str(clinResults[[1]],list.len=6)
# or listviewer::jsonedit(clinResults)
```

# Basic design

This package design is meant to have some similarities to the "hadleyverse"
approach of dplyr. Roughly, the functionality for finding and accessing files
and metadata can be divided into:

1. Simple query constructors based on GDC API endpoints.
2. A set of verbs that when applied, adjust filtering, field selection, and
faceting (fields for aggregation) and result in a new query object (an
endomorphism)
3. A set of verbs that take a query and return results from the GDC

In addition, there are exhiliary functions for asking the GDC API for
information about available and default fields, slicing BAM files, and
downloading actual data files.  Here is an overview of functionality[^1].


- Creating a query
    - `projects()`
    - `cases()`
    - `files()`
    - `annotations()`
- Manipulating a query
    - `filter()`
    - `facet()`
    - `select()`
- Introspection on the GDC API fields
    - `mapping()`
    - `available_fields()`
    - `default_fields()`
    - `grep_fields()`
    - `available_values()`
    - `available_expand()`
- Executing an API call to retrieve query results
    - `results()`
    - `count()`
    - `response()`
- Raw data file downloads
    - `gdcdata()`
    - `transfer()`
    - `gdc_client()`
- Summarizing and aggregating field values (faceting)
    - `aggregations()`
- Authentication
    - `gdc_token()`
- BAM file slicing
    - `slicing()`

[^1]: See individual function and methods documentation for specific details.


# Usage

There are two main classes of operations when working with the NCI GDC.

1. [Querying metadata and finding data files](#querying-metadata) (e.g., finding
all gene expression quantifications data files for all colon cancer patients).
2. [Transferring raw or processed data](#datafile-access-and-download) from the
GDC to another computer (e.g., downloading raw or processed data)

Both classes of operation are reviewed in detail in the following sections.

## Querying metadata

Vast amounts of metadata about cases (patients, basically), files, projects, and
so-called annotations are available via the NCI GDC API. Typically, one will
want to query metadata to either focus in on a set of files for download or
transfer *or* to perform so-called aggregations (pivot-tables, facets, similar
to the R `table()` functionality).

Querying metadata starts with [creating a "blank" query](#creating-a-query). One
will often then want to [`filter`](#filtering) the query to limit results prior
to [retrieving results](#retrieving-results). The GenomicDataCommons package has
[helper functions for listing fields](#fields-and-values) that are available for
filtering.

In addition to fetching results, the GDC API allows
[faceting, or aggregating,](#facets-and-aggregation), useful for compiling
reports, generating dashboards, or building user interfaces to GDC data (see GDC
web query interface for a non-R-based example).

### Creating a query

A query of the GDC starts its life in R. Queries follow the four metadata
endpoints available at the GDC.  In particular, there are four convenience
functions that each create `GDCQuery` objects (actually, specific subclasses of
`GDCQuery`):

- `projects()`
- `cases()`
- `files()`
- `annotations()`

```{r projectquery}
pquery = projects()
```

The `pquery` object is now an object of (S3) class, `GDCQuery` (and
`gdc_projects` and `list`). The object contains the following elements:

- fields: This is a character vector of the fields that will be returned when we
[retrieve data](#retrieving-results). If no fields are specified to, for
example, the `projects()` function, the default fields from the GDC are used
(see `default_fields()`)
- filters: This will contain results after calling the
[`filter()` method](#filtering) and will be used to filter results on
[retrieval](#retrieving-results).
- facets: A character vector of field names that will be used for
[aggregating data](#facets-and-aggregation) in a call to `aggregations()`.
- token: A character(1) token from the GDC. See
[the authentication section](#authentication) for details, but note that, in
general, the token is not necessary for metadata query and retrieval, only for
actual data download.

Looking at the actual object (get used to using `str()`!), note that the query
contains no results.

```{r pquery}
str(pquery)
```
### Retrieving results

[[ GDC pagination documentation ]](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#size-and-from)

[[ GDC sorting documentation ]](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#sort)

With a query object available, the next step is to retrieve results from the
GDC. The GenomicDataCommons package.  The most basic type of results we can get
is a simple `count()` of records available that satisfy the filter criteria.
Note that we have not set any filters, so a `count()` here will represent all
the project records publicly available at the GDC in the "default" archive"

```{r pquerycount}
pcount = count(pquery)
# or
pcount = pquery |> count()
pcount
```

The `results()` method will fetch actual results.

```{r pqueryresults}
presults = pquery |> results()
```
These results are
returned from the GDC in [JSON](http://www.json.org/) format and
converted into a (potentially nested) list in R. The `str()` method is useful
for taking a quick glimpse of the data.

```{r presultsstr}
str(presults)
```

A default of only 10 records are returned. We can use the `size` and `from`
arguments to `results()` to either page through results or to change the number
of results. Finally, there is a convenience method, `results_all()` that will
simply fetch all the available results given a query. Note that `results_all()`
may take a long time and return HUGE result sets if not used carefully. Use of a
combination of `count()` and `results()` to get a sense of the expected data
size is probably warranted before calling `results_all()`

```{r presultsall}
length(ids(presults))
presults = pquery |> results_all()
length(ids(presults))
# includes all records
length(ids(presults)) == count(pquery)
```

Extracting subsets of
results or manipulating the results into a more conventional R data
structure is not easily generalizable.  However,
the
[purrr](https://github.com/hadley/purrr),
[rlist](https://renkun.me/rlist/),
and [data.tree](https://cran.r-project.org/web/packages/data.tree/vignettes/data.tree.html) packages
are all potentially of interest for manipulating complex, nested list
structures. For viewing the results in an interactive viewer, consider the
[listviewer](https://github.com/timelyportfolio/listviewer) package.


### Fields and Values

[[ GDC `fields` documentation ]](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#fields)

Central to querying and retrieving data from the GDC is the ability to specify
which fields to return, filtering by fields and values, and faceting or
aggregating. The GenomicDataCommons package includes two simple functions,
`available_fields()` and `default_fields()`. Each can operate on a character(1)
endpoint name ("cases", "files", "annotations", or "projects") or a `GDCQuery`
object.

```{r defaultfields}
default_fields('files')
# The number of fields available for files endpoint
length(available_fields('files'))
# The first few fields available for files endpoint
head(available_fields('files'))
```

The fields to be returned by a query can be specified following a similar
paradigm to that of the dplyr package. The `select()` function is a verb that
resets the fields slot of a `GDCQuery`; note that this is not quite analogous to
the dplyr `select()` verb that limits from already-present fields. We
*completely replace* the fields when using `select()` on a `GDCQuery`.

```{r selectexample}
# Default fields here
qcases = cases()
qcases$fields
# set up query to use ALL available fields
# Note that checking of fields is done by select()
qcases = cases() |> GenomicDataCommons::select(available_fields('cases'))
head(qcases$fields)
```

Finding fields of interest is such a common operation that the
GenomicDataCommons includes the `grep_fields()` function.
See the appropriate help pages for details.

### Facets and aggregation

[[ GDC `facet` documentation ]](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#facets)

The GDC API offers a feature known as aggregation or faceting. By
specifying one or more fields (of appropriate type), the GDC can
return to us a count of the number of records matching each potential
value. This is similar to the R `table` method. Multiple fields can be
returned at once, but the GDC API does not have a cross-tabulation
feature; all aggregations are only on one field at a time. Results of
`aggregation()` calls come back as a list of data.frames (actually,
tibbles).

```{r aggexample}
# total number of files of a specific type
res = files() |> facet(c('type','data_type')) |> aggregations()
res$type
```

Using `aggregations()` is an also easy way to learn the contents of individual
fields and forms the basis for faceted search pages.

### Filtering

[[ GDC `filtering` documentation ]](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#filters-specifying-the-query)

The GenomicDataCommons package uses a form of non-standard evaluation to specify
R-like queries that are then translated into an R list. That R list is, upon
calling a method that fetches results from the GDC API, translated into the
appropriate JSON string. The R expression uses the formula interface as
suggested by Hadley Wickham in his [vignette on non-standard evaluation](https://cran.r-project.org/web/packages/dplyr/vignettes/nse.html)

> It’s best to use a formula because a formula captures both the expression to
evaluate and the environment where the evaluation occurs. This is important if
the expression is a mixture of variables in a data frame and objects in the
local environment [for example].

For the user, these details will not be too important except to note that a
filter expression must begin with a "~".

```{r allfilesunfiltered}
qfiles = files()
qfiles |> count() # all files
```
To limit the file type, we can refer back to the
[section on faceting](#facets-and-aggregation) to see the possible values for
the file field "type". For example, to filter file results to only
"gene_expression" files, we simply specify a filter.

```{r onlyGeneExpression}
qfiles = files() |> filter( type == 'gene_expression')
# here is what the filter looks like after translation
str(get_filter(qfiles))
```

What if we want to create a filter based on the project ('TCGA-OVCA', for
example)? Well, we have a couple of possible ways to discover available fields.
The first is based on base R functionality and some intuition.

```{r filtAvailFields}
grep('pro',available_fields('files'),value=TRUE) |> 
    head()
```

Interestingly, the project information is "nested" inside the case. We don't
need to know that detail other than to know that we now have a few potential
guesses for where our information might be in the files records.  We need to
know where because we need to construct the appropriate filter.

```{r filtProgramID}
files() |> 
    facet('cases.project.project_id') |> 
    aggregations() |> 
    head()
```

We note that `cases.project.project_id` looks like it is a good fit. We also
note that `TCGA-OV` is the correct project_id, not `TCGA-OVCA`. Note that
*unlike with dplyr and friends, the `filter()` method here **replaces** the
filter and does not build on any previous filters*.

```{r filtfinal}
qfiles = files() |>
    filter( cases.project.project_id == 'TCGA-OV' & type == 'gene_expression')
str(get_filter(qfiles))
qfiles |> count()
```

Asking for a `count()` of results given these new filter criteria gives `r
qfiles |> count()` results. Filters can be chained (or nested) to 
accomplish the same effect as multiple `&` conditionals. The `count()`
below is equivalent to the `&` filtering done above.

```{r filtChain}
qfiles2 = files() |>
    filter( cases.project.project_id == 'TCGA-OV') |> 
    filter( type == 'gene_expression') 
qfiles2 |> count()
(qfiles |> count()) == (qfiles2 |> count()) #TRUE
```



Generating a manifest for bulk downloads is as
simple as asking for the manifest from the current query.

```{r filtAndManifest}
manifest_df = qfiles |> manifest()
head(manifest_df)
```

Note that we might still not be quite there. Looking at filenames, there are
suspiciously named files that might include "FPKM", "FPKM-UQ", or "counts".
Another round of `grep` and `available_fields`, looking for "type" turned up
that the field "analysis.workflow_type" has the appropriate filter criteria.


```{r filterForSTARCounts}
qfiles = files() |> filter( ~ cases.project.project_id == 'TCGA-OV' &
                            type == 'gene_expression' &
                            access == "open" &
                            analysis.workflow_type == 'STAR - Counts')
manifest_df = qfiles |> manifest()
nrow(manifest_df)
```

The GDC Data Transfer Tool can be used (from R, `transfer()` or from the
command-line) to orchestrate high-performance, restartable transfers of all the
files in the manifest. See [the bulk downloads section](bulk-downloads) for
details.


## Authentication

[[ GDC authentication documentation ]](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#facets)

The GDC offers both "controlled-access" and "open" data. As of this
writing, only data stored as files is "controlled-access"; that is,
metadata accessible via the GDC is all "open" data and some files are
"open" and some are "controlled-access". Controlled-access data are
only available
after
[going through the process of obtaining access.](https://gdc.cancer.gov/access-data/obtaining-access-controlled-data)

After controlled-access to one or more datasets has been granted,
logging into the GDC web portal will allow you
to
[access a GDC authentication token](https://docs.gdc.cancer.gov/Data_Portal/Users_Guide/Authentication/#gdc-authentication-tokens),
which can be downloaded and then used to access available
controlled-access data via the GenomicDataCommons package.

The GenomicDataCommons uses authentication tokens only for downloading
data (see `transfer` and `gdcdata` documentation). The package
includes a helper function, `gdc_token`, that looks for the token to
be stored in one of three ways (resolved in this order):

1. As a string stored in the environment variable, `GDC_TOKEN`
2. As a file, stored in the file named by the environment variable,
   `GDC_TOKEN_FILE`
3. In a file in the user home directory, called `.gdc_token`

As a concrete example:

```{r authenNoRun, eval=FALSE}
token = gdc_token()
transfer(...,token=token)
# or
transfer(...,token=get_token())
```


## Datafile access and download

### Data downloads via the GDC API

The `gdcdata` function takes a character vector of one or more file
ids. A simple way of producing such a vector is to produce a
`manifest` data frame and then pass in the first column, which will
contain file ids.

```{r singlefileDL}
fnames = gdcdata(manifest_df$id[1:2],progress=FALSE)

```

Note that for controlled-access data, a
GDC [authentication token](#authentication) is required. Using the
`BiocParallel` package may be useful for downloading in parallel,
particularly for large numbers of smallish files.

### Bulk downloads

The bulk download functionality is only efficient (as of v1.2.0 of the
GDC Data Transfer Tool) for relatively large files, so use this
approach only when transferring BAM files or larger VCF files, for
example. Otherwise, consider using the approach shown above, perhaps
in parallel.

```{r bulkDL, eval=FALSE}
# Requires gcd_client command-line utility to be isntalled
# separately. 
fnames = gdcdata(manifest_df$id[3:10], access_method = 'client')
```


### BAM slicing

# Use Cases

## Cases

### How many cases are there per project_id?

```{r casesPerProject}
res = cases() |> facet("project.project_id") |> aggregations()
head(res)
library(ggplot2)
ggplot(res$project.project_id,aes(x = key, y = doc_count)) +
    geom_bar(stat='identity') +
    theme(axis.text.x = element_text(angle = 45, hjust = 1))
```

### How many cases are included in all TARGET projects?

```{r casesInTCGA}
cases() |> filter(~ project.program.name=='TARGET') |> count()
```

### How many cases are included in all TCGA projects?

```{r casesInTARGET}
cases() |> filter(~ project.program.name=='TCGA') |> count()
```

### What is the breakdown of sample types in TCGA-BRCA?

```{r casesTCGABRCASampleTypes}
# The need to do the "&" here is a requirement of the
# current version of the GDC API. I have filed a feature
# request to remove this requirement.
resp = cases() |> filter(~ project.project_id=='TCGA-BRCA' &
                              project.project_id=='TCGA-BRCA' ) |>
    facet('samples.sample_type') |> aggregations()
resp$samples.sample_type
```

### Fetch all samples in TCGA-BRCA that use "Solid Tissue" as a normal.

```{r casesTCGABRCASolidNormal}
# The need to do the "&" here is a requirement of the
# current version of the GDC API. I have filed a feature
# request to remove this requirement.
resp = cases() |> filter(~ project.project_id=='TCGA-BRCA' &
                              samples.sample_type=='Solid Tissue Normal') |>
    GenomicDataCommons::select(c(default_fields(cases()),'samples.sample_type')) |>
    response_all()
count(resp)
res = resp |> results()
str(res[1],list.len=6)
head(ids(resp))
```

### Get all TCGA case ids that are female

```{r casesFemaleTCGA}
cases() |>
  GenomicDataCommons::filter(~ project.program.name == 'TCGA' &
    "cases.demographic.gender" %in% "female") |>
      GenomicDataCommons::results(size = 4) |>
        ids()
```

### Get all TCGA-COAD case ids that are NOT female

```{r notFemaleTCGACOAD}
cases() |>
  GenomicDataCommons::filter(~ project.project_id == 'TCGA-COAD' &
    "cases.demographic.gender" %exclude% "female") |>
      GenomicDataCommons::results(size = 4) |>
        ids()
```

### Get all TCGA cases that are missing gender

```{r missingGenderTCGA}
cases() |>
  GenomicDataCommons::filter(~ project.program.name == 'TCGA' &
    missing("cases.demographic.gender")) |>
      GenomicDataCommons::results(size = 4) |>
        ids()
```

### Get all TCGA cases that are NOT missing gender

```{r notMissingGenderTCGA}
cases() |>
  GenomicDataCommons::filter(~ project.program.name == 'TCGA' &
    !missing("cases.demographic.gender")) |>
      GenomicDataCommons::results(size = 4) |>
        ids()
```


## Files

### How many of each type of file are available?

```{r filesVCFCount}
res = files() |> facet('type') |> aggregations()
res$type
ggplot(res$type,aes(x = key,y = doc_count)) + geom_bar(stat='identity') +
    theme(axis.text.x = element_text(angle = 45, hjust = 1))
```

### Find gene-level RNA-seq quantification files for GBM

```{r filesRNAseqGeneGBM}
q = files() |>
    GenomicDataCommons::select(available_fields('files')) |>
    filter(~ cases.project.project_id=='TCGA-GBM' &
               data_type=='Gene Expression Quantification')
q |> facet('analysis.workflow_type') |> aggregations()
# so need to add another filter
file_ids = q |> filter(~ cases.project.project_id=='TCGA-GBM' &
                            data_type=='Gene Expression Quantification' &
                            analysis.workflow_type == 'STAR - Counts') |>
    GenomicDataCommons::select('file_id') |>
    response_all() |>
    ids()
```

## Slicing

### Get all BAM file ids from TCGA-GBM

**I need to figure out how to do slicing reproducibly in a testing environment
and for vignette building**.

```{r filesRNAseqGeneGBMforBAM}
q = files() |>
    GenomicDataCommons::select(available_fields('files')) |>
    filter(~ cases.project.project_id == 'TCGA-GBM' &
               data_type == 'Aligned Reads' &
               experimental_strategy == 'RNA-Seq' &
               data_format == 'BAM')
file_ids = q |> response_all() |> ids()
```


```{r slicing10, eval=FALSE}
bamfile = slicing(file_ids[1],regions="chr12:6534405-6538375",token=gdc_token())
library(GenomicAlignments)
aligns = readGAlignments(bamfile)
```

# Troubleshooting

## SSL connection errors

* Symptom: Trying to connect to the API results in:
```
Error in curl::curl_fetch_memory(url, handle = handle) :
SSL connect error
```
* Possible solutions: The [issue
is that the GDC supports only recent security Transport Layer Security (TLS)](http://stackoverflow.com/a/42599546/459633),
so the only known fix is to upgrade the system `openssl` to version
1.0.1 or later.
    * [[Mac OS]](https://github.com/Bioconductor/GenomicDataCommons/issues/35#issuecomment-284233510),
    * [[Ubuntu]](http://askubuntu.com/a/434245)
    * [[Centos/RHEL]](https://www.liquidweb.com/kb/update-and-patch-openssl-for-the-ccs-injection-vulnerability/).
    After upgrading `openssl`, reinstall the R `curl` and `httr` packages.


# sessionInfo()

```{r sessionInfo}
sessionInfo()
```

# Developer notes

- The `S3` object-oriented programming paradigm is used.
- We have adopted a functional programming style with functions and methods that
often take an "object" as the first argument. This style lends itself to
pipeline-style programming.
- The GenomicDataCommons package uses the
[alternative request format (POST)](https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#alternative-request-format)
to allow very large request bodies.

---
title: "Working with simple somatic mutations"
author: "Sean Davis"
date: "`r format(Sys.Date(), '%A, %B %d, %Y')`"
always_allow_html: yes
output:
  BiocStyle::html_document:
    df_print: paged
    toc_float: true
abstract: >

vignette: >
  %\VignetteIndexEntry{Somatic Mutation Data}
  %\VignetteEngine{knitr::rmarkdown}
  %\VignetteEncoding{UTF-8}
---

# Background



# Workflow

```{r warning=FALSE,message=FALSE}
library(GenomicDataCommons)
library(tibble)
```

## Genes and gene details

```{r}
grep_fields('genes', 'symbol')
```
```{r}
head(available_values('genes','symbol'))
```



```{r}
tp53 = genes() |> 
  GenomicDataCommons::filter(symbol=='TP53') |> 
  results(size=10000) |> 
  as_tibble()
```


## ssms

```{r}
ssms() |> 
    GenomicDataCommons::filter(
      chromosome==paste0('chr',tp53$gene_chromosome[1]) &
        start_position > tp53$gene_start[1] & 
        end_position < tp53$gene_end[1]) |> 
    GenomicDataCommons::count()
```

```{r}
ssms() |> 
    GenomicDataCommons::filter(
      consequence.transcript.gene.symbol %in% c('TP53')) |> 
    GenomicDataCommons::count()
```

## convert to VRanges

```{r warning=FALSE,message=FALSE}
library(VariantAnnotation)
vars = ssms() |> 
    GenomicDataCommons::filter(
      consequence.transcript.gene.symbol %in% c('TP53')) |> 
    GenomicDataCommons::results_all() |>
    as_tibble()
```

```{r}
vr = VRanges(seqnames = vars$chromosome,
             ranges = IRanges(start=vars$start_position, width=1),
             ref = vars$reference_allele,
             alt = vars$tumor_allele)
```

```{r}
ssm_occurrences() |> 
    GenomicDataCommons::filter(
      ssm.consequence.transcript.gene.symbol %in% c('TP53')) |>
    GenomicDataCommons::count()
```

```{r}
var_samples = ssm_occurrences() |> 
    GenomicDataCommons::filter(
      ssm.consequence.transcript.gene.symbol %in% c('TP53')) |> 
    GenomicDataCommons::expand(c('case', 'ssm', 'case.project')) |>
    GenomicDataCommons::results_all() |> 
    as_tibble()
```

```{r}
table(var_samples$case$disease_type)
```

## OncoPrint

```{r}
fnames <- files() |>
  GenomicDataCommons::filter(
    cases.project.project_id=='TCGA-SKCM' &
      data_format=='maf' &
      data_type=='Masked Somatic Mutation' &
      analysis.workflow_type ==
        'Aliquot Ensemble Somatic Variant Merging and Masking'
  ) |>
  results(size = 1) |>
    ids() |>
      gdcdata()
```

```{r cache=TRUE}
library(maftools)
melanoma = read.maf(maf = fnames)
```

```{r}
maftools::oncoplot(melanoma)
```

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdcdata.R
\name{gdcdata}
\alias{gdcdata}
\title{Download GDC files}
\usage{
gdcdata(
  uuids,
  use_cached = TRUE,
  progress = interactive(),
  token = NULL,
  access_method = "api",
  transfer_args = character(),
  ...
)
}
\arguments{
\item{uuids}{character() of GDC file UUIDs.}

\item{use_cached}{logical(1) default TRUE indicating that,
if found in the cache, the file will not be downloaded
again. If FALSE, all supplied uuids will be re-downloaded.}

\item{progress}{logical(1) default TRUE in interactive sessions,
FALSE otherwise indicating whether a progress par should be
produced for each file download.}

\item{token}{(optional) character(1) security token allowing access
to restricted data. See
\url{https://gdc-docs.nci.nih.gov/API/Users_Guide/Authentication_and_Authorization/}.}

\item{access_method}{character(1), either 'api' or 'client'. See details.}

\item{transfer_args}{character(1), additional arguments to pass to
the gdc-client command line. See \code{\link{gdc_client}} and
\code{\link{transfer_help}} for details.}

\item{...}{further arguments passed to files}
}
\value{
a named vector with file uuids as the names and paths as
the value
}
\description{
Download one or more files from GDC. Files are downloaded using the
UUID and renamed to the file name on the remote system. By default,
neither the uuid nor the file name on the remote system can exist.
}
\details{
This function is appropriate for one or several files; for large
downloads use \code{\link{manifest}} to create a manifest for and
the GDC Data Transfer Tool.

When access_method is "api", the GDC "data" endpoint is the
    transfer mechanism used. The alternative access_method, "client", will
    utilize the \code{gdc-client} transfer tool, which must be
    downloaded separately and available. See
    \code{\link{gdc_client}} for details on specifying the location
    of the gdc-client executable.
}
\examples{
# get some example file uuids
uuids <- files() |>
    filter(~ access == 'open' & file_size < 100000) |>
    results(size = 3) |>
    ids()

# and get the data, placing it into the gdc_cache() directory
gdcdata(uuids, use_cached=TRUE)

}
\seealso{
\code{\link{manifest}} for downloading large data.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bulk_transfer.R
\name{gdc_client}
\alias{gdc_client}
\title{return gdc-client executable path}
\usage{
gdc_client()
}
\value{
character(1) the path to the gdc-client executable.
}
\description{
This function is a convenience function to 
find and return the path to the GDC Data Transfer
Tool executable assumed to be named 'gdc-client'. 
The assumption is that the appropriate version of the
GDC Data Transfer Tool is a separate download available
from \href{the GDC website}{https://gdc.cancer.gov/access-data/gdc-data-transfer-tool}
and as a backup from \href{on github}{https://github.com/NCI-GDC/gdc-client}.
}
\details{
The path is checked in the following order:
\enumerate{
\item an R option("gdc_client")
\item an environment variable GDC_CLIENT
\item from the search PATH
\item in the current working directory
}
}
\examples{
# this cannot run without first
# downloading the GDC Data Transfer Tool
gdc_client = try(gdc_client(),silent=TRUE)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fields.R
\name{grep_fields}
\alias{grep_fields}
\title{Find matching field names}
\usage{
grep_fields(entity, pattern, ..., value = TRUE)
}
\arguments{
\item{entity}{one of the available gdc entities ('files','cases',...)
against which to gather available fields for matching}

\item{pattern}{A regular expression that will be used
in a call to \code{\link{grep}}}

\item{...}{passed on to grep}

\item{value}{logical(1) whether to return values as opposed
to indices (passed along to grep)}
}
\value{
character() vector of field names matching
    \code{pattern}
}
\description{
This utility function allows quick text-based search of available
fields for using \code{\link{grep}}
}
\examples{
grep_fields('files','analysis')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/clinical.R
\name{gdc_clinical}
\alias{gdc_clinical}
\title{Get clinical information from GDC}
\usage{
gdc_clinical(case_ids, include_list_cols = FALSE)
}
\arguments{
\item{case_ids}{a character() vector of case_ids, typically from
"cases" query.}

\item{include_list_cols}{logical(1), whether to include list
columns in the "main" data.frame. These list columns have
values for aliquots, samples, etc. While these may be useful
for some situations, they are generally not that useful as
clinical annotations.}
}
\value{
A list of four data.frames:
\enumerate{
\item main, representing basic case identification and metadata
    (update date, etc.)
\item diagnoses
\item esposures
\item demographic
}
}
\description{
The NCI GDC has a complex data model that allows various studies to
supply numerous clinical and demographic data elements. However,
across all projects that enter the GDC, there are
similarities. This function returns four data.frames associated
with case_ids from the GDC.
}
\details{
Note that these data.frames can, in general, have different numbers
of rows (or even no rows at all). If one wishes to combine to
produce a single data.frame, using the approach of left joining to
the "main" data.frame will yield a useful combined data.frame. We
do not do that directly given the potential for 1:many
relationships. It is up to the user to determine what the best
approach is for any given dataset.
}
\examples{
case_ids = cases() |> results(size=10) |> ids()
clinical_data = gdc_clinical(case_ids)

# overview of clinical results
class(clinical_data)
names(clinical_data)
sapply(clinical_data, class)
sapply(clinical_data, nrow)

# available data
head(clinical_data$main)
head(clinical_data$demographic)
head(clinical_data$diagnoses)
head(clinical_data$exposures)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/gdc_token.R
\name{gdc_token}
\alias{gdc_token}
\title{return a gdc token from file or environment}
\usage{
gdc_token()
}
\value{
character(1) (invisibly, to protect against inadvertently printing) the GDC token.
}
\description{
The GDC requires an auth token for downloading
data that are "controlled access". For example, 
BAM files for human datasets, germline variant calls,
and SNP array raw data all are protected as "controlled
access". For these files, a GDC access token is required.
See the \href{details on the GDC authentication and token information}{https://docs.gdc.cancer.gov/Data_Portal/Users_Guide/Authentication/#gdc-authentication-tokens}.
Note that this function simply returns a string value. 
It is possible to keep the GDC token in a variable in R
or to pass a string directly to the appropriate parameter.
This function is simply a convenience function for alternative 
approaches to get a token from an environment variable
or a file.
}
\details{
This function will resolve locations of the GDC token in the 
following order:
\itemize{
\item{from the environment variable, \code{GDC_TOKEN}, expected to 
contain the token downloaded from the GDC as a string}
\item{using \code{readLines} to read a file named in the environment
variable, \code{GDC_TOKEN_FILE}}
\item{using \code{readLines} to read from a file called \code{.gdc_token} in the user's
home directory}
}
If all of these fail, this function will return an error.
}
\examples{
# This will not run before a GDC token
# is in place.  
token = try(gdc_token(),silent=TRUE)


}
\references{
\url{https://docs.gdc.cancer.gov/Data_Portal/Users_Guide/Cart/#gdc-authentication-tokens}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.R
\name{expand}
\alias{expand}
\alias{expand.GDCQuery}
\title{Set the \code{expand} parameter}
\usage{
expand(x, expand)

\method{expand}{GDCQuery}(x, expand)
}
\arguments{
\item{x}{the objects on which to set fields}

\item{expand}{a character vector specifying the fields}
}
\value{
A \code{\link{GDCQuery}} object, with the \code{expand}
member altered.
}
\description{
S3 generic to set GDCQuery expand parameter
}
\section{Methods (by class)}{
\itemize{
\item \code{expand(GDCQuery)}: set expand fields on a GDCQuery object

}}
\examples{
gProj = projects()
gProj$fields
head(available_fields(gProj))
default_fields(gProj)

gProj |>
  select(default_fields(gProj)[1:2]) |>
  response() |>
  str(max_level=2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{make_filter}
\alias{make_filter}
\title{Create NCI GDC filters for limiting GDC query results}
\usage{
make_filter(expr, available_fields)
}
\arguments{
\item{expr}{a lazy-wrapped expression or a formula RHS equivalent}

\item{available_fields}{a character vector of the
additional names that will be injected into the
filter evaluation environment}
}
\value{
a \code{list} that represents an R version
of the JSON that will ultimately be used in an
NCI GDC search or other query.
}
\description{
Searching the NCI GDC allows for complex filtering based
on logical operations and simple comparisons.  This function
facilitates writing such filter expressions in R-like syntax
with R code evaluation.
}
\details{
If used with available_fields, "bare" fields that are
named in the available_fields character vector can be used
in the filter expression without quotes.
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{response}
\alias{response}
\alias{GDCResponse}
\alias{response.GDCQuery}
\alias{response_all}
\title{Fetch \code{\link{GDCQuery}} metadata from GDC}
\usage{
response(x, ...)

\method{response}{GDCQuery}(x, from = 0, size = 10, ..., response_handler = jsonlite::fromJSON)

response_all(x, ...)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}

\item{...}{passed to httr (good for passing config info, etc.)}

\item{from}{integer index from which to start returning data}

\item{size}{number of records to return}

\item{response_handler}{a function that processes JSON (as text)
and returns an R object.  Default is \code{\link[jsonlite]{fromJSON}}.}
}
\value{
A \code{GDCResponse} object which is a list with the following
members:
\itemize{
\item{results}
\item{query}
\item{aggregations}
\item{pages}
}
}
\description{
Fetch \code{\link{GDCQuery}} metadata from GDC
}
\examples{

# basic class stuff
gCases = cases()
resp = response(gCases)
class(resp)
names(resp)

# And results from query
resp$results[[1]]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/query.R
\name{query}
\alias{query}
\alias{GDCQuery}
\alias{cases}
\alias{files}
\alias{projects}
\alias{annotations}
\alias{ssms}
\alias{ssm_occurrences}
\alias{cnvs}
\alias{cnv_occurrences}
\alias{genes}
\title{Start a query of GDC metadata}
\usage{
query(
  entity,
  filters = NULL,
  facets = NULL,
  expand = NULL,
  fields = default_fields(entity),
  ...
)

cases(...)

files(...)

projects(...)

annotations(...)

ssms(...)

ssm_occurrences(...)

cnvs(...)

cnv_occurrences(...)

genes(...)
}
\arguments{
\item{entity}{character vector, including one of the entities in .gdc_entities}

\item{filters}{a filter list, typically created using \code{\link{make_filter}}, or added
to an existing \code{GDCQuery} object using \code{\link{filter}}.}

\item{facets}{a character vector of facets for counting common values. 
See \code{\link{available_fields}}. In general, one will not specify this parameter
but will use \code{\link{facet}} instead.}

\item{expand}{a character vector of "expands" to include in returned data. See 
\code{\link{available_expand}}}

\item{fields}{a character vector of fields to return. See \code{\link{available_fields}}.
In general, one will not specify fields directly, but instead use \code{\link{select}}}

\item{...}{passed through to \code{\link{query}}}
}
\value{
An S3 object, the GDCQuery object. This is a list
with the following members.
\itemize{
\item{filters}
\item{facets}
\item{fields}
\item{expand}
\item{archive}
\item{token}
}
}
\description{
The basis for all functionality in this package
starts with constructing a query in R. The GDCQuery
object contains the filters, facets, and other
parameters that define the returned results. A token
is required for accessing certain datasets.
}
\section{Functions}{
\itemize{
\item \code{cases()}: convenience constructor for a GDCQuery for cases

\item \code{files()}: convenience contructor for a GDCQuery for files

\item \code{projects()}: convenience contructor for a GDCQuery for projects

\item \code{annotations()}: convenience contructor for a GDCQuery for annotations

\item \code{ssms()}: convenience contructor for a GDCQuery for ssms

\item \code{ssm_occurrences()}: convenience contructor for a GDCQuery for ssm_occurrences

\item \code{cnvs()}: convenience contructor for a GDCQuery for cnvs

\item \code{cnv_occurrences()}: convenience contructor for a GDCQuery for cnv_occurrences

\item \code{genes()}: convenience contructor for a GDCQuery for genes

}}
\examples{
qcases = query('cases')
# equivalent to:
qcases = cases()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fields.R
\name{field_description}
\alias{field_description}
\alias{field_description.GDCQuery}
\alias{field_description.character}
\title{S3 Generic that returns the field description text, if available}
\usage{
field_description(entity, field)

\method{field_description}{GDCQuery}(entity, field)

\method{field_description}{character}(entity, field)
}
\arguments{
\item{entity}{character(1) string ('cases','files','projects',
'annotations', etc.) or an subclass of \code{\link{GDCQuery}}.}

\item{field}{character(1), the name of the field that will be used to look
up the description.}
}
\value{
character(1) descriptive text or character(0) if no description
is available.
}
\description{
S3 Generic that returns the field description text, if available
}
\section{Methods (by class)}{
\itemize{
\item \code{field_description(GDCQuery)}: GDCQuery method

\item \code{field_description(character)}: character method

}}
\examples{
field_description('cases', 'annotations.category')
casesQuery = query('cases')
field_description(casesQuery, 'annotations.category')
field_description(cases(), 'annotations.category')

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/manifest.R
\name{write_manifest}
\alias{write_manifest}
\title{write a manifest data.frame to disk}
\usage{
write_manifest(manifest, destfile = tempfile())
}
\arguments{
\item{manifest}{A data.frame with five columns, typically
created by a call to \code{\link{manifest}}}

\item{destfile}{The filename for saving the manifest.}
}
\value{
character(1) the destination file name.
}
\description{
The \code{\link{manifest}} method creates a data.frame
that represents the data for a manifest file needed
by the GDC Data Transfer Tool. While the file format
is nothing special, this is a simple helper function
to write a manifest data.frame to disk. It returns
the path to which the file is written, so it can
be used "in-line" in a call to \code{\link{transfer}}.
}
\examples{
mf = files() |> manifest(size=10)
write_manifest(mf)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/GenomicDataCommons-package.R
\docType{package}
\name{GenomicDataCommons-package}
\alias{GenomicDataCommons}
\alias{GenomicDataCommons-package}
\title{GenomicDataCommons: A package for interfacing with the NCI GDC}
\description{
Programmatically access the NIH / NCI Genomic Data Commons RESTful service.
}
\section{finding data}{


\itemize{
\item{\code{\link{query}}}
\item{\code{\link{cases}}}
\item{\code{\link{projects}}}
\item{\code{\link{files}}}
\item{\code{\link{annotations}}}
\item{\code{\link{mapping}}}
}
}

\section{downloading data}{

data
}

\seealso{
Useful links:
\itemize{
  \item \url{https://bioconductor.org/packages/GenomicDataCommons}
  \item \url{http://github.com/Bioconductor/GenomicDataCommons}
  \item \url{http://bioconductor.github.io/GenomicDataCommons/}
  \item Report bugs at \url{https://github.com/Bioconductor/GenomicDataCommons/issues/new}
}

}
\author{
\strong{Maintainer}: Sean Davis \email{seandavi@gmail.com}

Authors:
\itemize{
  \item Martin Morgan \email{martin.morgan@roswellpark.org}
}

Other contributors:
\itemize{
  \item Marcel Ramos \email{marcel.ramos@sph.cuny.edu} [contributor]
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{count}
\alias{count}
\alias{count.GDCQuery}
\alias{count.GDCResponse}
\title{provide count of records in a \code{\link{GDCQuery}}}
\usage{
count(x, ...)

\method{count}{GDCQuery}(x, ...)

\method{count}{GDCResponse}(x, ...)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}

\item{...}{passed to httr (good for passing config info, etc.)}
}
\value{
integer(1) representing the count of records that will
 be returned by the current query
}
\description{
provide count of records in a \code{\link{GDCQuery}}
}
\section{Methods (by class)}{
\itemize{
\item \code{count(GDCQuery)}: 

\item \code{count(GDCResponse)}: 

}}
\examples{
# total number of projects
projects() |> count()

# total number of cases
cases() |> count()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{aggregations}
\alias{aggregations}
\alias{aggregations.GDCQuery}
\alias{aggregations.GDCResponse}
\title{aggregations}
\usage{
aggregations(x)

\method{aggregations}{GDCQuery}(x)

\method{aggregations}{GDCResponse}(x)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}
}
\value{
a \code{list} of \code{data.frame} with one
member for each requested facet. The data frames
each have two columns, key and doc_count.
}
\description{
aggregations
}
\section{Methods (by class)}{
\itemize{
\item \code{aggregations(GDCQuery)}: 

\item \code{aggregations(GDCResponse)}: 

}}
\examples{
# Number of each file type
res = files() |> facet(c('type','data_type')) |> aggregations()
res$type

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/facets.R
\name{facet}
\alias{facet}
\alias{get_facets}
\alias{get_facets.GDCQuery}
\title{Set facets for a \code{\link{GDCQuery}}}
\usage{
facet(x, facets)

get_facets(x)

\method{get_facets}{GDCQuery}(x)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}

\item{facets}{a character vector of fields that
will be used for forming aggregations (facets).
Default is to set facets for all default fields.
See \code{\link{default_fields}} for details}
}
\value{
returns a \code{\link{GDCQuery}} object,
with facets field updated.
}
\description{
Set facets for a \code{\link{GDCQuery}}

Get facets for a \code{\link{GDCQuery}}
}
\examples{
# create a new GDCQuery against the projects endpoint
gProj = projects()

# default facets are NULL
get_facets(gProj)

# set facets and save result
gProjFacet = facet(gProj)

# check facets
get_facets(gProjFacet)

# and get a response, noting that
# the aggregations list member contains
# tibbles for each facet
str(response(gProjFacet,size=2),max.level=2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/bulk_transfer.R
\name{transfer}
\alias{transfer}
\alias{gdc_client_version_validate}
\alias{transfer_help}
\title{Bulk data download}
\usage{
transfer(uuids, args = character(), token = NULL, overwrite = FALSE)

gdc_client_version_validate(valid_version = .GDC_COMPATIBLE_VERSION)

transfer_help()
}
\arguments{
\item{uuids}{character() vector of GDC file UUIDs}

\item{args}{character() vector specifying command-line arguments to
be passed to \code{gdc-client}. See \code{\link{transfer_help}} for
possible values. The arguments \code{--manifest}, \code{--dir},
and \code{--token-file} are determined by \code{manifest},
\code{destination_dir}, and \code{token}, respectively, and
should NOT be provided as elements of \code{args}.}

\item{token}{character(1) containing security
token allowing access to restricted data. See
\url{https://gdc-docs.nci.nih.gov/API/Users_Guide/Authentication_and_Authorization/}.
Note that the GDC transfer tool requires a file for data
transfer. Therefore, this token will be written to a temporary
file (with appropriate permissions set).}

\item{overwrite}{logical(1) default FALSE indicating whether
existing files with identical name should be over-written.}

\item{valid_version}{character(1) The last known version that works for the
current data release for which to validate against, not typically changed
by the end-user.}
}
\value{
character(1) directory path to which the files were
    downloaded.
}
\description{
The GDC maintains a special tool,
\href{the GDC Data Transfer Tool}{https://docs.gdc.cancer.gov/Data_Transfer_Tool/Users_Guide/Getting_Started/},
that enables high-performance, potentially parallel, and
resumable downloads. The Data Transfer Tool is an external
program that requires separate download. Due to recent changes in the
GDC API, the transfer function now validates the version of the `gdc-client`
to ensure reliable downloads.
}
\section{Functions}{
\itemize{
\item \code{gdc_client_version_validate()}: If you are using the 'client' option, your `gdc-client` should be
up-to-date (>= 1.3.0).

\item \code{transfer_help()}: 

}}
\examples{
\dontrun{
uuids = files() |> 
  filter(access == "open") |> 
  results() |>
  ids()
file_paths <- transfer(uuids)
file_paths
names(file_paths)
# and with authenication
# REQUIRES gdc_token 
# destination <- transfer(uuids,token=gdc_token())
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fields.R
\name{select}
\alias{select}
\alias{select.GDCQuery}
\title{S3 generic to set GDCQuery fields}
\usage{
select(x, fields)

\method{select}{GDCQuery}(x, fields)
}
\arguments{
\item{x}{the objects on which to set fields}

\item{fields}{a character vector specifying the fields}
}
\value{
A \code{\link{GDCQuery}} object, with the fields
member altered.
}
\description{
S3 generic to set GDCQuery fields
}
\section{Methods (by class)}{
\itemize{
\item \code{select(GDCQuery)}: set fields on a GDCQuery object

}}
\examples{
gProj = projects()
gProj$fields
head(available_fields(gProj))
default_fields(gProj)

gProj |>
  select(default_fields(gProj)[1:2]) |>
  response() |>
  str(max_level=2)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/constants.R
\name{endpoints}
\alias{endpoints}
\alias{parameters}
\title{Endpoints and Parameters}
\usage{
endpoints()

parameters()
}
\value{
\code{endpoints()} returns a character vector of possible
    endpoints.

\code{parameters()} returns a list of possible parameters
    and their default values.
}
\description{
\code{endpoints()} returns available endpoints.
}
\examples{
endpoints()
parameters()
}
\keyword{internal}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{results}
\alias{results}
\alias{results.GDCQuery}
\alias{results.GDCResponse}
\title{results}
\usage{
results(x, ...)

\method{results}{GDCQuery}(x, ...)

\method{results}{GDCResponse}(x, ...)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}

\item{...}{passed on to \code{\link{response}}}
}
\value{
A (typically nested) \code{list} of GDC records
}
\description{
results
}
\section{Methods (by class)}{
\itemize{
\item \code{results(GDCQuery)}: 

\item \code{results(GDCResponse)}: 

}}
\examples{
qcases = cases() |> results()
length(qcases)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/entity_name.R
\name{entity_name}
\alias{entity_name}
\alias{entity_name.GDCQuery}
\alias{entity_name.GDCResults}
\title{Get the entity name from a GDCQuery object}
\usage{
entity_name(x)

\method{entity_name}{GDCQuery}(x)

\method{entity_name}{GDCResults}(x)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}
}
\value{
character(1) name of an associated entity; one of
"cases", "files", "projects", "annotations".
}
\description{
An "entity" is simply one of the four medata endpoints.
\itemize{
\item{cases}
\item{projects}
\item{files}
\item{annotations}
}
All \code{\link{GDCQuery}} objects will have an entity name. This S3 method
is simply a utility accessor for those names.
}
\examples{
qcases = cases()
qprojects = projects()

entity_name(qcases)
entity_name(qprojects)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/mapping.R
\name{mapping}
\alias{mapping}
\title{Query GDC for available endpoint fields}
\usage{
mapping(endpoint)
}
\arguments{
\item{endpoint}{character(1) corresponding to endpoints for which
users may specify additional or alternative fields. Endpoints
include \dQuote{projects}, \dQuote{cases}, \dQuote{files}, and
\dQuote{annotations}.}
}
\value{
A data frame describing the field (field name), full (full
    data model name), type (data type), and four additional columns
    describing the "set" to which the fields belong--\dQuote{default},
    \dQuote{expand}, \dQuote{multi}, and \dQuote{nested}.
}
\description{
Query GDC for available endpoint fields
}
\examples{
map <- mapping("projects")
head(map)
# get only the "default" fields
subset(map,defaults)
# And get just the text names of the "default" fields
subset(map,defaults)$field

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fields.R
\name{available_values}
\alias{available_values}
\title{Find common values for a GDC field}
\usage{
available_values(entity, field)
}
\arguments{
\item{entity}{character(1), a GDC entity ("cases", "files", "annotations", "projects")}

\item{field}{character(1), a field that is present in the entity record}
}
\value{
character vector of the top 100 (or fewer) most frequent
    values for a the given field
}
\description{
Find common values for a GDC field
}
\examples{
available_values('files','cases.project.project_id')[1:5]

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/caching.R
\name{gdc_cache}
\alias{gdc_cache}
\alias{gdc_set_cache}
\title{Work with gdc cache directory}
\usage{
gdc_cache()

gdc_set_cache(
  directory = rappdirs::app_dir(appname = "GenomicDataCommons")$cache(),
  verbose = TRUE,
  create_without_asking = !interactive()
)
}
\arguments{
\item{directory}{character(1) directory path, will be created
recursively if not present.}

\item{verbose}{logical(1) whether or not to message the location of
the cache directory after creation.}

\item{create_without_asking}{logical(1) specifying whether to allow
the function to create the cache directory without asking the
user first. In an interactive session, if the cache directory
does not exist, the user will be prompted before creation.}
}
\value{
character(1) directory path that serves as
the base directory for GenomicDataCommons downloads.

the created directory (invisibly)
}
\description{
The GenomicDataCommons package will cache downloaded
files to minimize network and allow for
offline work. These functions are used to create a cache directory
if one does not exist, set a global option, and query that
option. The cache directory will default to the user "cache"
directory according to specifications in
\code{\link[rappdirs]{app_dir}}. However, the user may want to set
this to another direcotory with more or higher performance
storage.
}
\details{
The cache structure is currently just a directory with each file
    being represented by a path constructed as:
    CACHEDIR/UUID/FILENAME. The cached files can be manipulated
    using standard file system commands (removing, finding,
    etc.). In this sense, the cache sytem is minimalist in design.
}
\section{Functions}{
\itemize{
\item \code{gdc_set_cache()}: (Re)set the GenomicDataCommons cache
directory

}}
\examples{
gdc_cache()
\dontrun{
gdc_set_cache(getwd())
}

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readDNAcopy.R
\name{readDNAcopy}
\alias{readDNAcopy}
\title{Read DNAcopy results into GRanges object}
\usage{
readDNAcopy(fname, ...)
}
\arguments{
\item{fname}{The path to a DNAcopy-like file.}

\item{...}{passed to \code{\link[readr]{read_tsv}}}
}
\value{
a \code{\link[GenomicRanges]{GRanges}} object
}
\description{
Read DNAcopy results into GRanges object
}
\examples{
fname = system.file(package='GenomicDataCommons',
                    'extdata/dnacopy.tsv.gz')
dnac = readDNAcopy(fname)
class(dnac)
length(dnac)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/fields.R
\name{available_fields}
\alias{available_fields}
\alias{available_fields.GDCQuery}
\alias{available_fields.character}
\title{S3 Generic to return all GDC fields}
\usage{
available_fields(x)

\method{available_fields}{GDCQuery}(x)

\method{available_fields}{character}(x)
}
\arguments{
\item{x}{A character(1) string ('cases','files','projects',
'annotations') or an subclass of \code{\link{GDCQuery}}.}
}
\value{
a character vector of the default fields
}
\description{
S3 Generic to return all GDC fields
}
\section{Methods (by class)}{
\itemize{
\item \code{available_fields(GDCQuery)}: GDCQuery method

\item \code{available_fields(character)}: character method

}}
\examples{
available_fields('projects')
projQuery = query('projects')
available_fields(projQuery)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids.R
\name{ids}
\alias{ids}
\alias{ids.GDCManifest}
\alias{ids.GDCQuery}
\alias{ids.GDCResults}
\alias{ids.GDCResponse}
\title{Get the ids associated with a GDC query or response}
\usage{
ids(x)

\method{ids}{GDCManifest}(x)

\method{ids}{GDCQuery}(x)

\method{ids}{GDCResults}(x)

\method{ids}{GDCResponse}(x)
}
\arguments{
\item{x}{A \code{\link{GDCQuery}} or \code{\link{GDCResponse}} object}
}
\value{
a character vector of all the entity ids
}
\description{
The GDC assigns ids (in the form of uuids) to objects in its database. Those
ids can be used for relationships, searching on the website, and as
unique ids.  All
}
\examples{
# use with a GDC query, in this case for "cases"
ids(cases() |> filter(~ project.project_id == "TCGA-CHOL"))
# also works for responses
ids(response(files()))
# and results
ids(results(cases()))


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/readHTSeqFile.R
\name{readHTSeqFile}
\alias{readHTSeqFile}
\title{Read a single htseq-counts result file.}
\usage{
readHTSeqFile(fname, samplename = "sample", ...)
}
\arguments{
\item{fname}{character(1), the path of the htseq-count file.}

\item{samplename}{character(1), the name of the sample. This will
become the name of the second column on the resulting
\code{data.frame}, making for easier merging if necessary.}

\item{...}{passed to \code{\link[readr]{read_tsv})}}
}
\value{
a two-column data frame
}
\description{
The htseq package is used extensively to count reads
relative to regions (see 
\url{http://www-huber.embl.de/HTSeq/doc/counting.html}).
The output of htseq-count is a simple two-column table
that includes features in column 1 and counts in column 2.
This function simply reads in the data from one such file
and assigns column names.
}
\examples{
fname = system.file(package='GenomicDataCommons',
                    'extdata/example.htseq.counts.gz')
dat = readHTSeqFile(fname)
head(dat)

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/status.R
\name{status}
\alias{status}
\title{Query the GDC for current status}
\usage{
status(version = NULL)
}
\arguments{
\item{version}{(optional) character(1) version of GDC}
}
\value{
List describing current status.
}
\description{
Query the GDC for current status
}
\examples{
status()

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/manifest.R
\name{manifest}
\alias{manifest}
\alias{manifest.gdc_files}
\alias{manifest.GDCfilesResponse}
\alias{manifest.GDCcasesResponse}
\title{Prepare GDC manifest file for bulk download}
\usage{
manifest(x, from = 0, size = count(x), ...)

\method{manifest}{gdc_files}(x, from = 0, size = count(x), ...)

\method{manifest}{GDCfilesResponse}(x, from = 0, size = count(x), ...)

\method{manifest}{GDCcasesResponse}(x, from = 0, size = count(x), ...)
}
\arguments{
\item{x}{An \code{\link{GDCQuery}} object of subclass "gdc_files" or "gdc_cases".}

\item{from}{Record number from which to start when returning the manifest.}

\item{size}{The total number of records to return.  Default 
will return the usually desirable full set of records.}

\item{...}{passed to \code{\link[httr]{PUT}}.}
}
\value{
A \code{\link[tibble]{tibble}}, also of type "gdc_manifest", with five columns:
\itemize{
\item{id}
\item{filename}
\item{md5}
\item{size}
\item{state}
}
}
\description{
The \code{manifest} function/method creates a manifest of files to be downloaded
using the GDC Data Transfer Tool. There are methods for
creating manifest data frames from \code{\link{GDCQuery}} objects
that contain file information ("cases" and "files" queries).
}
\section{Methods (by class)}{
\itemize{
\item \code{manifest(gdc_files)}: 

\item \code{manifest(GDCfilesResponse)}: 

\item \code{manifest(GDCcasesResponse)}: 

}}
\examples{
gFiles = files()
shortManifest = gFiles |> manifest(size=10)
head(shortManifest,n=3)


}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/ids.R
\name{id_field}
\alias{id_field}
\alias{id_field.GDCQuery}
\alias{id_field.GDCResults}
\title{get the name of the id field}
\usage{
id_field(x)

\method{id_field}{GDCQuery}(x)

\method{id_field}{GDCResults}(x)
}
\arguments{
\item{x}{An object representing the query or results 
of an entity from the GDC ("cases", "files", "annotations", "projects")}
}
\value{
character(1) such as "case_id", "file_id", etc.
}
\description{
In many places in the GenomicDataCommons package,
the entity ids are stored in a column or a vector
with a specific name that corresponds to the field name 
at the GDC. The format is the entity name (singular) "_id".
This generic simply returns that name from a given object.
}
\section{Methods (by class)}{
\itemize{
\item \code{id_field(GDCQuery)}: GDCQuery method

\item \code{id_field(GDCResults)}: GDCResults method

}}
\examples{
id_field(cases())

}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/slicing.R
\name{slicing}
\alias{slicing}
\title{Query GDC for data slices}
\usage{
slicing(
  uuid,
  regions,
  symbols,
  destination = file.path(tempdir(), paste0(uuid, ".bam")),
  overwrite = FALSE,
  progress = interactive(),
  token = gdc_token()
)
}
\arguments{
\item{uuid}{character(1) identifying the BAM file resource}

\item{regions}{character() vector describing chromosomal regions,
e.g., \code{c("chr1", "chr2:10000", "chr3:10000-20000")} (all
of chromosome 1, chromosome 2 from position 10000 to the end,
chromosome 3 from 10000 to 20000).}

\item{symbols}{character() vector of gencode gene symbols, e.g.,
\code{c("BRCA1", "PTEN")}}

\item{destination}{character(1) default \code{tempfile()} file path
for BAM file slice}

\item{overwrite}{logical(1) default FALSE can destination be
overwritten?}

\item{progress}{logical(1) default \code{interactive()} should a
progress bar be used?}

\item{token}{character(1) security token allowing access to
restricted data. Almost all BAM data is restricted, so a token is
usually required. See
\url{https://docs.gdc.cancer.gov/Data/Data_Security/Data_Security/#authentication-tokens}.}
}
\value{
character(1) destination to the downloaded BAM file
}
\description{
This function returns a BAM file representing reads overlapping
regions specified either as chromosomal regions or as gencode gene
symbols.
}
\details{
This function uses the Genomic Data Commons "slicing" API
    to get portions of a BAM file specified either using "regions"
    or using HGNC gene symbols.
}
\examples{
\dontrun{
 slicing("df80679e-c4d3-487b-934c-fcc782e5d46e",
        regions="chr17:75000000-76000000",
        token=gdc_token())

# Get 10 BAM files.
bamfiles = files() |> 
           filter(data_format=='BAM') |>
           results(size=10) |> ids()

# Current alignments at the GDC are to GRCh38
library('TxDb.Hsapiens.UCSC.hg38.knownGene')
all_genes = genes(TxDb.Hsapiens.UCSC.hg38.knownGene)

first3genes = all_genes[1:3]
# remove strand info
strand(first3genes) = '*'

# We can get our regions easily now
as.character(first3genes)

# Use parallel downloads to speed processing
library(BiocParallel)
register(MulticoreParam())

fnames = bplapply(bamfiles, slicing, overwrite = TRUE,
                regions=as.character(first3genes))

# 10 BAM files
fnames

library(GenomicAlignments)
lapply(unlist(fnames), readGAlignments)

}
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/expand.R
\name{available_expand}
\alias{available_expand}
\alias{available_expand.character}
\alias{available_expand.GDCQuery}
\title{Return valid values for "expand"}
\usage{
available_expand(entity)

\method{available_expand}{character}(entity)

\method{available_expand}{GDCQuery}(entity)
}
\arguments{
\item{entity}{Either a \code{\link{GDCQuery}} object
or a character(1) specifying a GDC entity ('cases', 'files',
'annotations', 'projects')}
}
\value{
A character vector
}
\description{
The GDC allows a shorthand for specifying groups
of fields to be returned by the metadata queries.
These can be specified in a \code{\link{select}}
method call to easily supply groups of fields.
}
\examples{
head(available_expand('files'))

}
\seealso{
See \url{https://docs.gdc.cancer.gov/API/Users_Guide/Search_and_Retrieval/#expand}
for details
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/filters.R
\name{filtering}
\alias{filtering}
\alias{filter}
\alias{filter.GDCQuery}
\alias{get_filter}
\alias{get_filter.GDCQuery}
\title{Manipulating GDCQuery filters}
\usage{
filter(x, expr)

\method{filter}{GDCQuery}(x, expr)

get_filter(x)

\method{get_filter}{GDCQuery}(x)
}
\arguments{
\item{x}{the object on which to set the filter list
member}

\item{expr}{a filter expression in the form of
the right hand side of a formula, where bare names
(without quotes) are allowed if they are available
fields associated with the GDCQuery object, \code{x}}
}
\value{
A \code{\link{GDCQuery}} object with the filter
field replaced by specified filter expression
}
\description{
Manipulating GDCQuery filters

The \code{filter} is simply a safe accessor for
the filter element in \code{\link{GDCQuery}} objects.

The \code{get_filter} is simply a safe accessor for
the filter element in \code{\link{GDCQuery}} objects.
}
\examples{
# make a GDCQuery object to start
#
# Projects
#
pQuery = projects()

# check for the default fields
# so that we can use one of them to build a filter
default_fields(pQuery)
pQuery = filter(pQuery,~ project_id == 'TCGA-LUAC')
get_filter(pQuery)

#
# Files
#
fQuery = files()
default_fields(fQuery)

fQuery = filter(fQuery,~ data_format == 'VCF')
# OR
# with recent GenomicDataCommons versions:
#   no "~" needed
fQuery = filter(fQuery, data_format == 'VCF')

get_filter(fQuery)

fQuery = filter(fQuery,~ data_format == 'VCF'
                & experimental_strategy == 'WXS'
                & type == 'simple_somatic_mutation')

files() |> filter(~ data_format == 'VCF'
                   & experimental_strategy=='WXS'
                   & type == 'simple_somatic_mutation') |> count()
                   
                   
files() |> filter( data_format == 'VCF'
                   & experimental_strategy=='WXS'
                   & type == 'simple_somatic_mutation') |> count()

# Filters may be chained for the 
# equivalent query
# 
# When chained, filters are combined with logical AND

files() |>
  filter(~ data_format == 'VCF') |>
  filter(~ experimental_strategy == 'WXS') |>
  filter(~ type == 'simple_somatic_mutation') |>
  count()

# OR

files() |>
  filter( data_format == 'VCF') |>
  filter( experimental_strategy == 'WXS') |>
  filter( type == 'simple_somatic_mutation') |>
  count()

# Use str() to get a cleaner picture
str(get_filter(fQuery))
}

% Generated by roxygen2: do not edit by hand
% Please edit documentation in R/response.R
\name{results_all}
\alias{results_all}
\alias{results_all.GDCQuery}
\alias{results_all.GDCResponse}
\title{results_all}
\usage{
results_all(x)

\method{results_all}{GDCQuery}(x)

\method{results_all}{GDCResponse}(x)
}
\arguments{
\item{x}{a \code{\link{GDCQuery}} object}
}
\value{
A (typically nested) \code{list} of GDC records
}
\description{
results_all
}
\section{Methods (by class)}{
\itemize{
\item \code{results_all(GDCQuery)}: 

\item \code{results_all(GDCResponse)}: 

}}
\examples{
# details of all available projects
projResults = projects() |> results_all()
length(projResults)
count(projects())


}
